/* 
 *
 * This program generates a static calltree of all the functions listed in the modules provided
 *
 * Run with no arguments for the full 'usage', although typically 
 *
 * 	4glpc -t WRITE somemod1.4gl
 * 	4glpc -t WRITE somemod2.4gl
 *
 * 	fglcalltree somemod1 somemod2
 *
 * The output generated is : 
 *
 *     calltree.xml  - XML file which can be viewed using the calltree viewer - shows full calltracing including all conditionals needed for a function to be called
 *
 *     calltree.dot  - a 'graphviz' output file which can be processed by 'dot' to generate a graphical calltree (best used with an 'calltree.ignore' file (see later)
 *
 *     calltree.unl  - an Unload format file which can be loaded into a database to show what functions are defined and whether they are in fact used
 *			This should load into a table with the schema : 
 *
 *				programname CHAR(64),
 *				modulename  CHAR(64),
 *				lineno INTEGER,
 *				functionname CHAR(64),
 *				used CHAR(1)
 *                      the 'programname' will be taken from the A4GL_PROGNAME environment setting (which is set automatically in A4GL IDE makefiles), or
 *                      defaults to 'PROGRAM' - which is used as a placeholer and can be subsequently updated in SQL..
 *
 *
 * This program will read a file called calltree.ignore - which should contain a list of all the functions which are not 'interesting' - these are typically library functions
 * where we dont really need to list, or the functions which they themselves call.
 * This can make the "graphs" generated by processing calltree.dot much smaller and therefore more readable..
 *
 * This program can also read an Aubit4GL "$CFUNCSFILE" to determine which functions are external 'C' functions
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "commands.h"
#include "a4gl_4glc_int.h"
#include "lint.h"
#include "linearise.h"
#include "expr_munging.h"
#include "a4gl_API_sqlparsecmd.h"

#define MODE_TRY 0
#define MODE_BUY 1

static int get_first_lineno(struct s_commands *c) ;
static int get_last_lineno(struct s_commands *c) ;
FILE *output = 0;
char *last_mod="<not set>";
int last_line=0;
static int indent = 0;
char currfunc[2000]="";
char currmod[2000]="";
static void add_assignments_for_parameters(expr_str_list *params,char *module,int lineno) ;
static int cache_expression (char *s, expr_str ** ptr, int mode);
static int cache_expression_list (char *s, struct expr_str_list *srclist, int mode);
extern int yylineno;
char *lint_module = 0;
int inIf=0;
static int isLineCalled(char *module, int line) ;
//static void dump_crud(void) ;
static void calltree_map_sql(char *buff,char *module, int line) ;
static void load_boltons (char *fname);
//static void calltree_addmap(int type, char *data);
static int add_calltree_calls (char *s, s_commands * func_commands, int mode);
static void calltree_map_insert_delete_update(char *main_statement, char *table, struct expr_str *where_clause, char *module, int line) ;
static void add_file_operation(char *operation, expr_str *file, char *module,int line) ;
static void calltree_map_select_stmt(char *main_statement, struct s_select *select,  char *module, int line) ;
static char *variable_value_string (char *var,int useName) ;
static void calltree_map_value_list(char *main_statement,struct s_select_list_item_list *list, char*module, int line) ;
static void calltree_map_value(char *main_statement,struct s_select_list_item *item, char*module, int line) ;
static char * guess_sql_stmt(struct struct_prepare_cmd *p,char *module, int line,int processed) ;
static void add_variable_value_strings (char *var, char *value) ;
static void clr_variable_values(void) ;
static int find_function (char *s);
char *decode_rb (enum report_blocks a);
char *endTarget=0;
static int ignore_user_function(char *name) ;
int ignLibFunc=1;
static int run_calltree (char *s);
static void add_variable_value(variable_usage *u, expr_str *val) ;
static char * evaluate_expr (expr_str * e,int useName);
static int process_cmd(struct command *cmd,int mode) ;
static void dec_inIf(void) ;
static void inc_inIf(void) ;
int ifBlocks[100];
int inc4GL = 1;			/* Include 4gl in the output */
int incProg = 1;
int AlwaysIncludeAllConditionals=1;
char top_level_function[200]="MAIN";
int groupByModule=0;

static void add_symbol(char *name,char*mod,int line,char *type,char *operation) ;
static void add_variable(expr_str *v, char *module, int lineno, char *operation,expr_str *value) ;
static char *get_form_display(char *s);

static void log_special_calls(struct s_expr_function_call *fcall);

enum e_mapset {
        MAPSET_NONE=0,
        MAPSET_CRUD_SELECT,
        MAPSET_CRUD_UPDATE,
        MAPSET_CRUD_DELETE,
        MAPSET_CRUD_INSERT,
        MAPSET_CRUD_DYNAMIC,
        MAPSET_CRUD_OTHER,
        MAPSET_ENV,
        MAPSET_RUN,
        MAPSET_FORM,
        MAPSET_FILES,


        MAPSET_LAST
};


struct {
        char **lines;
        int nlines;
} map_crud[100];

static void calltree_addmap (enum e_mapset mapset, char *fmt, char *module,int line) ;
struct s_select *subq_stmts[2000];
//struct selects *subq_selects[2000];
int stmts_cnt=0;

enum e_mapset curr_mapset[100];
int curr_mapset_cnt=0;


static void init_mapsets(void) {
int a;
        for (a=0;a<100;a++) {
                map_crud[a].lines=0;
                map_crud[a].nlines=0;
        }
}



static void dump_mapset_to_crud(char *tag, enum e_mapset e) {
int a;
        FPRINTF(output,"      <%s>\n",tag);

        for (a=0;a<map_crud[e].nlines;a++) {
                FPRINTF(output,"         %s",map_crud[e].lines[a]);
                free(map_crud[e].lines[a]);
        }

        FPRINTF(output,"      </%s>\n",tag);

        free(map_crud[e].lines);
        map_crud[e].lines=0;
        map_crud[e].nlines=0;
}


static void dump_mapset(enum e_mapset e) {


        switch (e) {
                case MAPSET_NONE :              dump_mapset_to_crud("Impossibles",e); break;
                //case MAPSET_UI:                 dump_mapset_to_crud("UIs",e); break;
                //case MAPSET_RUN:                dump_mapset_to_crud("RUNS",e); break;
                case MAPSET_CRUD_SELECT:        dump_mapset_to_crud("SELECTS",e); break;
                case MAPSET_CRUD_UPDATE:        dump_mapset_to_crud("UPDATES",e); break;
                case MAPSET_CRUD_DELETE:        dump_mapset_to_crud("DELETES",e); break;
                case MAPSET_CRUD_INSERT:        dump_mapset_to_crud("INSERTS",e); break;
                //case MAPSET_CRUD_CREATE_TEMP:   dump_mapset_to_crud("TEMPTABLES",e); break;
                case MAPSET_CRUD_OTHER:         dump_mapset_to_crud("OTHERSQLS",e); break;
                //case MAPSET_OTHER:         dump_mapset_to_crud("OTHERSQLS",e); break;
                case MAPSET_CRUD_DYNAMIC:       dump_mapset_to_crud("DYNAMICSQLS",e); break;
                //case MAPSET_UI_OPENFORM:        dump_mapset_to_crud("FORMS",e); break;
                //case MAPSET_CALL:               dump_mapset_to_crud("CALLS",e); break;
                //case MAPSET_EVENTS:             dump_mapset_to_crud("EVENTS",e); break;
                //case MAPSET_DISPLAY_AT:       dump_mapset_to_crud("DISPLATATS",e); break;
                //case MAPSET_LIBCALL_WORKFLOW:   dump_mapset_to_crud("WORKFLOWCALLS",e); break;
                //case MAPSET_LIBCALL_CORR:       dump_mapset_to_crud("CORRESPONDANCECALLS",e); break;
                //case MAPSET_LIBCALL_AWB:        dump_mapset_to_crud("AWBCALLS",e); break;
                //case MAPSET_LIBCALL:            dump_mapset_to_crud("OTHERLIBCALLS",e); break;
                case MAPSET_ENV:                dump_mapset_to_crud("ENVIRONMENTVARS",e); break;
                case MAPSET_RUN:                dump_mapset_to_crud("RUN",e); break;
                case MAPSET_FORM:               dump_mapset_to_crud("FORMS",e); break;
                case MAPSET_FILES:              dump_mapset_to_crud("FILES",e); break;

                case MAPSET_LAST: break;
                //default :       A4GL_assertion(1,"Unexpected");
        }
}

static void dump_mapsets(void) {
int a;
        for (a=0;a<100;a++) {
                if (map_crud[a].nlines) {
                        dump_mapset(a);
                }
        }

}


extern module_definition this_module;
//int expr_datatype (struct expr_str *p);


#define PROTO_FUNCTION 		1
#define PROTO_REPORT 		2
#define PROTO_PDF_REPORT 	3

struct module_definition *current_mod;
struct s_function_prototype
{
  int proto_type;
  char *pname;
  int nparams;
  int *param_dtypes;
  int nreturns;
  int *return_dtypes;
};

struct s_nodes
{
  char *function;
  int lineno;
  char *module;
  char *calls;
  int checked;
	char type;
};

struct s_nodes *nodes = 0;
int nodescnt;

struct function_calls
{
  char *conditional;
  int nchildren;
  struct function_calls *children;

};


struct function
{
  char *module;
  int module_no;
  int line;
  int lastline;
  char *function;
  char f_or_r;
  struct function_calls *calls;
  void *ptr;
  int called;
  int callsEndTarget;
  int isInLibrary;

  char *whenever_error_func;
  char *whenever_any_error_func;
  char *whenever_sql_error_func;
  int whenever_error_func_line;
  int whenever_any_error_func_line;
  int whenever_sql_error_func_line;
  s_call_list *call_list;
  variable_list *local_variables;

};



struct function *functions = 0;
int functions_cnt = 0;




struct s_function_prototype *fprototypes = 0;

struct s_function_prototype *fprototypes_boltons = 0;
int nboltons = 0;
//static int is_bolton_function (char *s);

expr_str_list *expr_cache = 0;

int simpleGraph = 1;
int printAllFuncs = 0;

static int
hasNode (char *fname, char *calls,int lineno)
{
  int a;
  for (a = 0; a < nodescnt; a++) 
    {


  	if (simpleGraph)  {
      		if (strcmp (nodes[a].function, fname) == 0 && strcmp (nodes[a].calls, calls) == 0) return 1;
	} else {
      		if ((lineno==nodes[a].lineno || lineno==-1) && strcmp (nodes[a].function, fname) == 0 && strcmp (nodes[a].calls, calls) == 0) return 1;
	}
    }
  return 0;
}



static void
addNode (char *fname, char *calls,char *module, int line,int lastLine, char *type)
{

  if (endTarget) {
	int fid;
	// Lets skip anything that isn't on our callpath 
	fid=find_function(calls); if (fid!=-1) { if (functions[fid].callsEndTarget==0) return; }
	fid=find_function(fname); if (fid!=-1) { if (functions[fid].callsEndTarget==0) return; }
  }

  if (!hasNode (fname, calls,line))
    {
      nodescnt++;
      nodes = realloc (nodes, sizeof (struct s_nodes) * nodescnt);
      nodes[nodescnt - 1].function = strdup (fname);
      nodes[nodescnt - 1].calls = strdup (calls);
      nodes[nodescnt - 1].lineno = line;
      nodes[nodescnt - 1].module = module; /* This is only used when 'calls' = NODE_FUNC_DEFINED */
      nodes[nodescnt - 1].type=type[0];
      nodes[nodescnt - 1].checked=0;
    }
}

#define NODE_FUNC_DEFINED "**DEFINED**"

static void
addFunction (char *fname, char *module, int line,int lastline, char *type)

{
  addNode (fname, NODE_FUNC_DEFINED,module,line,lastline,type);
}


/* 
Create a new function definition for any function which are *called* but not defined
These are most likely to be C functions ...

*/
static void check_for_undefined_functions(void) {
	int a;
	for (a=0;a<nodescnt;a++) {
      		if (!hasNode (nodes[a].calls, NODE_FUNC_DEFINED, -1) || printAllFuncs) {
			addNode (nodes[a].calls, NODE_FUNC_DEFINED,"undefined",0,0,"U");
		}
	}
}


static void check_for_orphaned_functions(char *topfuncname) {
	int a;
	int topfuncno=-1;

	
	for (a=0;a<nodescnt;a++) {
		if (strcmp(topfuncname,nodes[a].function)==0) {
			if (strcmp(nodes[a].calls,NODE_FUNC_DEFINED)==0) {
				if (nodes[a].checked) return; /* Its already been checked */
				topfuncno=a;
			}
		}
	}


	if (topfuncno==-1) return; // Couldn't find it..

//printf("Checking for orphaned functions from :%s\n", topfuncname);
	nodes[topfuncno].checked=1; /* We've checked it now... */

	for (a=0;a<nodescnt;a++) {
		if (strcmp(topfuncname,nodes[a].function)==0) {
			if (strcmp(nodes[a].calls,NODE_FUNC_DEFINED)==0) {
				continue;
			}
			nodes[a].checked=1;
//printf("%s calls %s ? \n", topfuncname, nodes[a].calls);
			check_for_orphaned_functions(nodes[a].calls);
		}
	}
//printf("Done Checking for orphaned functions from :%s\n", topfuncname);
}

static char *get_prog_name(void) {
char *ptr;
ptr=acl_getenv_not_set_as_0("A4GL_PROGNAME");
if (ptr ==0) {
	return "PROGRAM";
}
return ptr;
}


struct s_group {
	char *modName;
	int nFuncs;
	char **funcnames;
	char **outstrs;
	int *used;
};

struct s_group *moduleGroups=0;
int nmoduleGroups=0;

static void addToGroup(int modNo,char*funcname,char *outstr) {
	int a;
	if (moduleGroups[modNo].nFuncs) {
		for (a=0;a<moduleGroups[modNo].nFuncs;a++)  {
			if (strcmp(moduleGroups[modNo].funcnames[a],funcname)==0) return; // Already there...
		}
	}
	moduleGroups[modNo].nFuncs++;
	moduleGroups[modNo].funcnames=realloc(moduleGroups[modNo].funcnames, sizeof(moduleGroups[modNo].funcnames[0])*moduleGroups[modNo].nFuncs);
	moduleGroups[modNo].funcnames[moduleGroups[modNo].nFuncs-1]=strdup(funcname);

	moduleGroups[modNo].used=realloc(moduleGroups[modNo].used, sizeof(moduleGroups[modNo].used[0])*moduleGroups[modNo].nFuncs);
	moduleGroups[modNo].used[moduleGroups[modNo].nFuncs-1]=0;

	moduleGroups[modNo].outstrs=realloc(moduleGroups[modNo].outstrs, sizeof(moduleGroups[modNo].outstrs[0])*moduleGroups[modNo].nFuncs);
	moduleGroups[modNo].outstrs[moduleGroups[modNo].nFuncs-1]=strdup(outstr);
}


static void add_group(char *funcname,char *modname,char *outstr) {
int a;
	if (moduleGroups) {
		for (a=0;a<nmoduleGroups;a++) {
			if (strcmp(moduleGroups[a].modName,modname)==0) {
				addToGroup(a,funcname, outstr);
				return;
			}
		}
	}
	nmoduleGroups++;
	moduleGroups=realloc(moduleGroups,sizeof(moduleGroups[0])*nmoduleGroups);
	moduleGroups[nmoduleGroups-1].modName=strdup(modname);
	moduleGroups[nmoduleGroups-1].nFuncs=0;
	moduleGroups[nmoduleGroups-1].funcnames=0;
	moduleGroups[nmoduleGroups-1].outstrs=0;
	moduleGroups[nmoduleGroups-1].used=0;

	addToGroup(nmoduleGroups-1,funcname,outstr);
}


static void set_used(char *funcname) {
int modNo;
int funcNo;
for (modNo=0;modNo<nmoduleGroups;modNo++) {
	for (funcNo=0;funcNo<moduleGroups[modNo].nFuncs;funcNo++) {
		if (strcmp(funcname,moduleGroups[modNo].funcnames[funcNo])==0) {
				moduleGroups[modNo].used[funcNo]++;
		}
	}
}
}

static void print_groupings(FILE *dot_output) {
int modNo;
int funcNo;


for (modNo=0;modNo<nmoduleGroups;modNo++) {

	if (groupByModule) {
		fprintf(dot_output,"\nsubgraph cluster_%s {\n", moduleGroups[modNo].modName);
		fprintf(dot_output," label = \"%s.4gl\";\n", moduleGroups[modNo].modName);
		fprintf(dot_output," style=filled; color = lightgrey;\n");
	}

	
	//fprintf(dot_output," node [style=filled,color=white];", moduleGroups[modNo].modName);
	for (funcNo=0;funcNo<moduleGroups[modNo].nFuncs;funcNo++) {

		if (moduleGroups[modNo].used[funcNo]==0) continue; // Ignore anything that isn't called - when -Gonly

		fprintf(dot_output,"	%s;\n",moduleGroups[modNo].outstrs[funcNo]);
	}


	if (groupByModule) {
		fprintf(dot_output,"}\n");
	}
}
fprintf(dot_output,"\n\n");

}


static int getModuleNo_forFunction(char *s) {
int modNo;
int funcNo;
for (modNo=0;modNo<nmoduleGroups;modNo++) {
	for (funcNo=0;funcNo<moduleGroups[modNo].nFuncs;funcNo++) {
		if (strcmp(s,moduleGroups[modNo].funcnames[funcNo])==0) {
			return modNo;
		}
	}
}
return -1;
}


static int isSameModule(char *f,char *c) {
	int mod_func;
	int mod_calls;
	mod_func=getModuleNo_forFunction(f);
	mod_calls=getModuleNo_forFunction(c);
	if (mod_func==mod_calls) return 1;
	return 0;
}


static void
printDot (void)
{
  FILE *dot_output = 0;
  FILE *unl_output = 0;
  char *mappedModName = 0;
char outStr[2000];
  int a;



  dot_output = fopen ("calltree.dot", "w");
  unl_output = fopen ("calltree.unl", "w");

  if (dot_output == 0)
    {
      printf ("Unable to open output file (calltree.dot)\n");
      exit (2);
    }

  fprintf (dot_output, "digraph G { // process with 'dot' - eg :   dot -o calltree.gif -Tgif calltree.dot\n");
  fprintf (dot_output, "rankdir=LR;\nratio=fill;\nsplines=polyline;\noverlap=vpsc;\n");

  check_for_undefined_functions ();
  check_for_orphaned_functions (top_level_function);


  for (a = 0; a < nodescnt; a++)
    {
      //int fid;
      //fid = find_function (nodes[a].function);


      if (strcmp (nodes[a].calls, NODE_FUNC_DEFINED) == 0)
	{
	  // Its our function definition placeholder...
	  //
	  mappedModName = nodes[a].module;
	  if (acl_getenv_not_set_as_0 ("V4GL"))
	    {
	      if (A4GL_strstartswith (mappedModName, acl_getenv_not_set_as_0 ("V4GL")))
		{
		  mappedModName += strlen (acl_getenv_not_set_as_0 ("V4GL"));
		  while (mappedModName[0] == '/')
		    mappedModName++;
		}
	    }

	  if (nodes[a].checked == 0)
	    {
	      if (strcmp (nodes[a].module, "undefined") != 0)
		{
		  fprintf (unl_output, "%s|%s|%d|%s|N\n", get_prog_name (), mappedModName, nodes[a].lineno, nodes[a].function);
		}
	      else
		{

		  if (strcmp (nodes[a].function, NODE_FUNC_DEFINED) != 0)
		    {
		      //if (is_bolton_function(nodes[a].function)==-1) {
		      fprintf (unl_output, "%s|missing|0|%s|M\n", get_prog_name (), nodes[a].function);
		      //}
		    }
		}
	      continue;
	    }

	  //if (fid==-1) continue;
	  //if (functions[fid].called==0) continue;

	  switch (nodes[a].type)
	    {
	    case 'M':		/* Main */
	      sprintf (outStr, "\"%s\" [ shape=record, label=< <table border=\"1\"><tr><td colspan=\"2\" bgcolor=\"#30ff30\">%s</td></tr><tr><td>%s</td><td>%d</td></tr></table> > ]\n", nodes[a].function, nodes[a].function, nodes[a].module, nodes[a].lineno);
	      fprintf (unl_output, "%s|%s|%d|%s|Y\n", get_prog_name (), mappedModName, nodes[a].lineno, nodes[a].function);
	      add_group( nodes[a].function,  nodes[a].module,outStr);
	      break;

	    case 'F':		/* Function */
	      sprintf (outStr,
		       "\"%s\" [ shape=record, label=< <table border=\"1\"><tr><td colspan=\"2\" bgcolor=\"#c0f0c0\">%s</td></tr><tr><td>%s</td><td>%d</td></tr></table> > ]\n",
		       nodes[a].function, nodes[a].function, nodes[a].module, nodes[a].lineno);
			add_group( nodes[a].function,  nodes[a].module,outStr);
	      fprintf (unl_output, "%s|%s|%d|%s|Y\n", get_prog_name (), mappedModName, nodes[a].lineno, nodes[a].function);
	      break;

	    case 'R':		/* Report */
	      sprintf (outStr,
		       "\"%s\" [ shape=record, label=< <table border=\"1\"><tr><td colspan=\"2\" bgcolor=\"#c0c0f0\">%s</td></tr><tr><td>%s</td><td>%d</td></tr></table> > ]\n",
		       nodes[a].function, nodes[a].function, nodes[a].module, nodes[a].lineno);
			add_group( nodes[a].function,  nodes[a].module,outStr);
	      fprintf (unl_output, "%s|%s|%d|%s|Y\n", get_prog_name (), mappedModName, nodes[a].lineno, nodes[a].function);
	      break;
	    case 'U':		/* Undefined */
	      if (printAllFuncs)
		{
		  sprintf (outStr,
			   "\"%s\" [ shape=record, label=< <table border=\"1\"><tr><td colspan=\"2\" bgcolor=\"#f0f0f0\">%s</td></tr><tr><td>%s</td><td>%d</td></tr></table> > ]\n",
			   nodes[a].function, nodes[a].function, nodes[a].module, nodes[a].lineno);
			add_group( nodes[a].function,  "Unknown",outStr);
		}
	      fprintf (unl_output, "%s|missing|0|%s|M\n", get_prog_name (), nodes[a].function);
	      break;
	    }
	}
    }






  for (a = 0; a < nodescnt; a++)
    {

      if (strcmp (nodes[a].calls, NODE_FUNC_DEFINED) == 0)
	continue;		// ignore this one - its just a placeholder...

      if (nodes[a].checked == 0)
	continue;


      if (hasNode (nodes[a].calls, NODE_FUNC_DEFINED, -1) || printAllFuncs)
	{

	  if (ignore_user_function (nodes[a].calls))
	    continue;


	  if (groupByModule==2) {
		if (isSameModule(nodes[a].function, nodes[a].calls)) {
			continue;
		}
	  }

	set_used(nodes[a].function);
	set_used(nodes[a].calls);
	}
    }



  print_groupings(dot_output);


  for (a = 0; a < nodescnt; a++)
    {

      if (strcmp (nodes[a].calls, NODE_FUNC_DEFINED) == 0)
	continue;		// ignore this one - its just a placeholder...

      if (nodes[a].checked == 0)
	continue;


      if (hasNode (nodes[a].calls, NODE_FUNC_DEFINED, -1) || printAllFuncs)
	{

	  if (ignore_user_function (nodes[a].calls))
	    continue;


	  if (groupByModule==2) {
		if (isSameModule(nodes[a].function, nodes[a].calls)) {
			continue;
		}
	  }

	  if (simpleGraph)
	    {
	      fprintf (dot_output, "\"%s\" -> \"%s\"\n", nodes[a].function, nodes[a].calls);
	    }
	  else
	    {
	      fprintf (dot_output, "\"%s\" -> \"%s\" [ label=\" Line:%d\" ]\n", nodes[a].function, nodes[a].calls, nodes[a].lineno);
	    }
	}
    }

  fprintf (dot_output, "}\n");
  fclose (dot_output);
  fclose (unl_output);
}


char *whenever_error_func = NULL;
char *whenever_any_error_func = NULL;
char *whenever_sql_error_func = NULL;

int whenever_error_func_line = 0;
int whenever_any_error_func_line = 0;
int whenever_sql_error_func_line = 0;


static void init_whenever (void)
{
  whenever_error_func = NULL;
  whenever_any_error_func = NULL;
  whenever_sql_error_func = NULL;

  whenever_error_func_line = 0;
  whenever_any_error_func_line = 0;
  whenever_sql_error_func_line = 0;
}



static void set_whenever_for_function(struct function *f) {

  if (f->whenever_error_func_line>whenever_error_func_line) { // The WHENEVER must have been done in between the functions..
 	whenever_error_func = f->whenever_error_func;
  }

  if (f->whenever_any_error_func_line> whenever_any_error_func_line) {
  	whenever_any_error_func = f->whenever_any_error_func;
  }

  if (f->whenever_sql_error_func_line> whenever_sql_error_func_line) {
  	whenever_sql_error_func = f->whenever_sql_error_func;
  }

}


static void
set_whenever (struct command *c)
{
  int whencode;
  char *whento;
  A4GL_assertion (c->cmd_data.type != E_CMD_WHENEVER_CMD, "Expecting a WHENEVER command");

  whencode = c->cmd_data.command_data_u.whenever_cmd.whencode;
  whento = c->cmd_data.command_data_u.whenever_cmd.whento;

  if (whencode & WHEN_ERROR)    { whenever_error_func = NULL; whenever_error_func_line = c->lineno; }
  if (whencode & WHEN_SQLERROR) { whenever_sql_error_func = NULL; whenever_sql_error_func_line = c->lineno; }
  if (whencode & WHEN_ANYERROR) { whenever_any_error_func = NULL; whenever_any_error_func_line = c->lineno; }


  if (whencode == WHEN_ERROR + WHEN_CALL)
    {
      whenever_error_func = whento;
      whenever_error_func_line = c->lineno;
    }

  if (whencode == WHEN_SQLERROR + WHEN_CALL)
    {
      whenever_sql_error_func = whento;
      whenever_sql_error_func_line = c->lineno;
    }

  if (whencode == WHEN_ANYERROR + WHEN_CALL)
    {
      whenever_any_error_func = whento;
      whenever_any_error_func_line = c->lineno;
    }

}



//static int promoteable(int a, int b) ;

//static void set_expr_cache(expr_str_list *n) { expr_cache=n; }

expr_str *
get_cached_expr_datatype (int n)
{
  A4GL_assertion (expr_cache == 0, "No expression cache");
  if (n > expr_cache->list.list_len || n < 0)
    {
      A4GL_assertion (1, "Expression cache subscript out of range");
    }
  return expr_cache->list.list_val[n];
}



char *
dtype_as_string (int dtype)
{
  if (dtype == 90)
    {
      return "Boolean";
    }
  switch (dtype & DTYPE_MASK)
    {
    case DTYPE_CHAR:
      return "char";
    case DTYPE_SMINT:
      return "smallint";
    case DTYPE_INT:
      return "integer";
    case DTYPE_FLOAT:
      return "float";
    case DTYPE_SMFLOAT:
      return "smallfloat";
    case DTYPE_DECIMAL:
      return "decimal";
    case DTYPE_SERIAL:
      return "serial";
    case DTYPE_DATE:
      return "date";
    case DTYPE_MONEY:
      return "money";
    case DTYPE_NULL:
      return "null";
    case DTYPE_DTIME:
      return "datetime";
    case DTYPE_BYTE:
      return "byte";
    case DTYPE_TEXT:
      return "text";
    case DTYPE_VCHAR:
      return "varchar";
    case DTYPE_INTERVAL:
      return "interval";
    case DTYPE_NCHAR:
      return "nchar";
    }
  return "Unknown";

}





int
local_is_valid_vname (struct variable *v, enum e_scope scope)
{
  char *nm;
  nm = v->names.names.names_val[0].name;

  if (strlen (nm) < 4)
    return 0;			// too short for sn_
  if (nm[2] != '_')
    return 0;

  if (nm[1] == 'v')
    {
      return 1;
    }

  switch (v->var_data.variable_type)
    {
    case VARIABLE_TYPE_SIMPLE:
      switch (v->var_data.variable_data_u.v_simple.datatype & DTYPE_MASK)
	{
	case DTYPE_CHAR:
	  if (nm[1] != 'c')
	    return 0;
	  else
	    return 1;
	case DTYPE_SMINT:
	  if (nm[1] != 'b' && nm[1] != 's')
	    return 0;
	  else
	    return 1;
	case DTYPE_SERIAL:
	case DTYPE_INT:
	  if (nm[1] != 'n')
	    return 0;
	  else
	    return 1;
	case DTYPE_DATE:
	  if (nm[1] != 'd')
	    return 0;
	  else
	    return 1;
	case DTYPE_DECIMAL:
	  if (nm[1] != 'l')
	    return 0;
	  else
	    return 1;
	case DTYPE_INTERVAL:
	  if (nm[1] != 'i')
	    return 0;
	  else
	    return 1;

	}
      // No specific rule - so must be ok...
      return 1;


	case VARIABLE_TYPE_USERDTYPE:
		return 1;

    case VARIABLE_TYPE_RECORD:
      if (nm[1] != 'r')
	return 0;
      else
	return 1;



    case VARIABLE_TYPE_ASSOC:
    case VARIABLE_TYPE_CONSTANT:
    case VARIABLE_TYPE_FUNCTION_DECLARE:
    case VARIABLE_TYPE_OBJECT:
	case VARIABLE_TYPE_TYPE_DECLARATION:
      return 1;


      //case VARIABLE_TYPE_ARRAY:
      //if (nm[1]!='a') return 0; else return 1;
    }

  return 1;
}



void
check_module (struct module_definition *d)
{
  //struct s_commands *all_cmds ;

 linearise_commands (0, 0);

}
























static int
system_function_dtype (char *funcname)
{

// Dont return anything
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_add_keymap") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_closeiem") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_dump_screen") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_load_datatype") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_run_gui") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_set_page") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_show_help") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_dump_screen") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_fetchiem") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_get_curr_height") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_get_curr_width") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_get_user") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_libhelp_showhelp") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_openiem") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_random") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_read_pipe") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_set_color") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_setenv") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_scr_size") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_getkey_wait") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_keysetlabel") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddeconnect") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddeexecute") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddefinish") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddefinishall") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddegeterror") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddepeek") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "ddepoke") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "err_print") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "err_quit") == 0)
    return -1;

  if (A4GL_isyes(acl_getenv("IGNORE_ERRLOG"))) ;
  else {
  	if (A4GL_aubit_strcasecmp (funcname, "errorlog") == 0) return -1;
	}

  if (A4GL_aubit_strcasecmp (funcname, "fgl_dialog_setbuffer") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_dialog_setcurrline") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_dialog_setkeylabel") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_prtscr") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_set_arrline") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_set_scrline") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_setkeylabel") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_strtosend") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_winmessage") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_drawbox") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "set_count") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "showhelp") == 0)
    return -1;

  if (A4GL_isyes(acl_getenv("IGNORE_ERRLOG"))) ;
  else {
  	if (A4GL_aubit_strcasecmp (funcname, "startlog") == 0) return -1;
	}


  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_send_to_ui") == 0)
    return -1;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_settitle") == 0)
    return -1;

  if (A4GL_aubit_strcasecmp (funcname, "fgl_fglgui") == 0)
    return 2;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_winbutton") == 0)
    return 2;


// INTs
  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_sendfile_to_ui") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "length") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "weekday") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "year") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "arr_count") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "arr_curr") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "day") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "length") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "month") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_winmessage") == 0)
    return DTYPE_INT; 

  if (A4GL_aubit_strcasecmp (funcname, "winexec") == 0)
    return DTYPE_SMINT;


  if (A4GL_aubit_strcasecmp (funcname, "mdy") == 0)
    return DTYPE_DATE;

  if (A4GL_aubit_strcasecmp (funcname, "a4gl_get_info") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_get_page") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "a4gl_get_ui_mode") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "arg_val") == 0)
    return DTYPE_CHAR;



  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_byte_as_str") == 0)
    return DTYPE_CHAR;



  if (A4GL_aubit_strcasecmp (funcname, "ascii") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "date") == 0)
    return DTYPE_DATE;
  if (A4GL_aubit_strcasecmp (funcname, "dbms_dialect") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "downshift") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "err_get") == 0)
    return DTYPE_CHAR;

  if (A4GL_aubit_strcasecmp (funcname, "aclfgl_expand_env_vars_in_cmdline") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_buffertouched") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_compare") == 0)
    return DTYPE_INT;

  if (A4GL_aubit_strcasecmp (funcname, "fgl_dialog_getbuffer") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_dialog_getfieldname") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_getenv") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_getkey") == 0)
    return DTYPE_CHAR;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_keyval") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_lastkey") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "fgl_round") == 0)
    return DTYPE_INT;


  if (A4GL_aubit_strcasecmp (funcname, "ord") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "num_args") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "scr_line") == 0)
    return DTYPE_INT;
  if (A4GL_aubit_strcasecmp (funcname, "upshift") == 0)
    return DTYPE_CHAR;



  if (A4GL_aubit_strcasecmp (funcname, "abs") == 0)
    return DTYPE_FLOAT;
  if (A4GL_aubit_strcasecmp (funcname, "sqrt") == 0)
    return DTYPE_FLOAT;

  return -2;
}



static int ignore_user_function(char *name) {
static char *names[2000];
static int n=0;

if (name==NULL) { // load it up..
	FILE *f;
  if (acl_getenv_not_set_as_0("V4GL")) {
     // Preload these as ignirable- as we dont want to count these hwne V4GL is set
     // (Ventas)
     names[n++]=strdup("let_textvar");
     names[n++]=strdup("suchen_case_total");
     names[n++]=strdup("let_textvar_kunde");
  }

char *calltree_ignore="calltree.ignore";
if (acl_getenv_not_set_as_0("CALLTREEIGNORE")) {
	calltree_ignore=acl_getenv_not_set_as_0("CALLTREEIGNORE");
}
	f=fopen(calltree_ignore,"r");
	if (f) {
		char buff[256];
		while (1) {
			fgets(buff,255,f);
			A4GL_trim_nl(buff);
			names[n++]=strdup(buff);
			if (feof(f)) break;
		}
	}
	return 1;
}

if (n) {
	int a;
	for (a=0;a<n;a++) {
		if (A4GL_aubit_strcasecmp(names[a],name)==0) {
			return 1;
		}
	}
}


if (ignLibFunc) {
	int a;
	a=find_function(name);
	if (a!=-1)  {
		if (functions[a].isInLibrary) return 1;
	}
}

return 0;

}

static int
system_function (char *funcname)
{
  if (system_function_dtype (funcname) != -2)
    {
      return 1;
    }
  if (ignore_user_function(funcname)) return 1;
 
  return 0;
}






static void
add_function (int module_no, char *module, int line, int lastline, char *fname, char forr, void *ptr,int isInLibrary,  s_call_list * call_list, variable_list *vlist)
{
  functions_cnt++;
  functions = realloc (functions, sizeof (struct function) * functions_cnt);
  functions[functions_cnt - 1].module = module;
  functions[functions_cnt - 1].module_no = module_no;
  functions[functions_cnt - 1].line = line;
  functions[functions_cnt - 1].lastline = lastline;
  functions[functions_cnt - 1].function = fname;
  functions[functions_cnt - 1].f_or_r = forr;
  functions[functions_cnt - 1].calls = 0;
  functions[functions_cnt - 1].ptr = ptr;
  functions[functions_cnt - 1].called = 0;
  functions[functions_cnt - 1].callsEndTarget = 0;
  functions[functions_cnt - 1].isInLibrary=isInLibrary;
  functions[functions_cnt - 1].call_list = call_list;
  functions[functions_cnt-1].local_variables=vlist;



  functions[functions_cnt - 1].whenever_error_func = whenever_error_func;
  functions[functions_cnt - 1].whenever_any_error_func = whenever_any_error_func;
  functions[functions_cnt - 1].whenever_sql_error_func = whenever_sql_error_func;
  functions[functions_cnt - 1].whenever_error_func_line = whenever_error_func_line;
  functions[functions_cnt - 1].whenever_any_error_func_line = whenever_any_error_func_line;
  functions[functions_cnt - 1].whenever_sql_error_func_line = whenever_sql_error_func_line;

  //printf ("Added %s in %s.%d\n", fname, module, line);
}

int callstack_cnt = 0;
char **callstack = 0;


/*
static void
add_to_callstack (char *s)
{
  callstack_cnt++;
  callstack = realloc (callstack, sizeof (char *) * callstack_cnt);
  callstack[callstack_cnt - 1] = s;
  if (callstack_cnt > 100)
    {
      printf ("Callstack too deep\n");
    }
}
*/


/*
static int
isoncallstack (char *s)
{
  int a;
  for (a = 0; a < callstack_cnt; a++)
    {
      if (strcmp (s, callstack[a]) == 0)
	return 1;
    }
  return 0;
}

static void
pop_callstack (void)
{
  callstack_cnt--;
}
*/

static int find_function (char *s)
{
  int a;
  for (a = 0; a < functions_cnt; a++)
    {
      if (strcmp (s, functions[a].function) == 0)
	return a;
    }
  return -1;
}

/*
static int
print_spaces (int n)
{
  int a;
  if (n <= 0)
    return 0;
  for (a = 0; a < n; a++)
    printf ("   ");

  return 0;
}
*/

static void
print_indent (void)
{
  int a;
  for (a = 0; a < indent; a++)
    {
      fprintf (output, "   ");
    }
}

/*
static int is_end_target(char *s) {
	// If nothing is set - it must be the end target...
	if (endTarget==0) return 1;

		printf("Is end target %s\n",s);
	if (strcmp(s,endTarget)==0) {
		printf("yez!\n");
		return 1;
	}
	printf("No\n");
	return 0;
}
*/

static int
cache_expression (char *sxx, expr_str ** ptr, int mode)
{
  expr_str *expr;
  if (ptr == 0)
    return 0;
  expr = *ptr;


  if (expr == 0)
    return 0;


  switch (expr->expr_type)
    {
    case ET_EXPR_NOT:
    case ET_EXPR_NEG:
    case ET_EXPR_OP_CLIP:
    case ET_EXPR_UPSHIFT:
    case ET_EXPR_DOWNSHIFT:
    case ET_EXPR_ASCII:
    case ET_EXPR_OP_ISNULL:
    case ET_EXPR_OP_ISNOTNULL:
    case ET_EXPR_YEAR_FUNC:
    case ET_EXPR_DATE_FUNC:
    case ET_EXPR_MONTH_FUNC:
    case ET_EXPR_DAY_FUNC:
    case ET_EXPR_TIME_FUNC:
    case ET_EXPR_OP_YEAR:
    case ET_EXPR_OP_MONTH:
    case ET_EXPR_OP_DAY:
    case ET_EXPR_OP_HOUR:
    case ET_EXPR_OP_MINUTE:
    case ET_EXPR_OP_SECOND:
      return cache_expression ("", &expr->expr_str_u.expr_expr, mode);

    case ET_EXPR_OP_MATCHES:
    case ET_EXPR_OP_LIKE:
    case ET_EXPR_OP_NOT_LIKE:
    case ET_EXPR_OP_NOT_MATCHES:
    case ET_EXPR_OP_ADD:
    case ET_EXPR_OP_SUB:
    case ET_EXPR_OP_DIV:
    case ET_EXPR_OP_MULT:
    case ET_EXPR_OP_LESS_THAN:
    case ET_EXPR_OP_GREATER_THAN:
    case ET_EXPR_OP_EQUAL:
    case ET_EXPR_OP_NOT_EQUAL:
    case ET_EXPR_OP_LESS_THAN_EQ:
    case ET_EXPR_OP_GREATER_THAN_EQ:
    case ET_EXPR_OP_OR:
    case ET_EXPR_OP_AND:
    case ET_EXPR_OP_USING:
    case ET_EXPR_OP_MOD:
    case ET_EXPR_OP_POWER:
    case ET_EXPR_OP_CONCAT:
      return cache_expression ("", &expr->expr_str_u.expr_op->left, mode) + cache_expression ("", &expr->expr_str_u.expr_op->right,
											      mode);

    case ET_EXPR_VARIABLE_USAGE:
	add_variable(expr,last_mod,last_line,"USE",expr);
	break;

    default:
      break;

    }

  if (expr->expr_type == ET_EXPR_FCALL)
    {
	int cnt=0;
      //int a;
      struct expr_str_list *params;
      // Got a function call - add it to the stack..
      if (mode == MODE_BUY)
	{
	
	  if (strcmp(expr->expr_str_u.expr_function_call->functionname,"fgl_getenv")==0) {
		char buff[2000];
		sprintf(buff,"<ENV NAME=\"%s\" MODULE=\"%s\" LINE=\"%d\" />\n", evaluate_expr(expr->expr_str_u.expr_function_call->parameters->list.list_val[0],1), last_mod,last_line);
		calltree_addmap(MAPSET_ENV ,buff,last_mod,last_line);

		}

	  log_special_calls(expr->expr_str_u.expr_function_call);
	

	  if (!system_function (expr->expr_str_u.expr_function_call->functionname))
	    {
		  addNode (currfunc, expr->expr_str_u.expr_function_call->functionname,"",expr->expr_str_u.expr_function_call->line,0,
"C");


	      print_indent ();
	      fprintf (output, "<CALLS FUNCTIONNAME='%s' LINE=\"%d\"/>\n", expr->expr_str_u.expr_function_call->functionname,
		       expr->expr_str_u.expr_function_call->line);
	    }
	}
      params = expr->expr_str_u.expr_function_call->parameters;
      if (params) {
	cnt+= cache_expression_list ("", params, mode);
	}

      if (!system_function (expr->expr_str_u.expr_function_call->functionname))
	{
		cnt++;
        }
	return cnt;
    }

  return 0;
}


static int
cache_expression_list (char *s, struct expr_str_list *srclist, int mode)
{
  int a;
  int cnt = 0;
  if (srclist == 0)
    return 0;
  if (srclist->list.list_len == 0)
    return 0;

  for (a = 0; a < srclist->list.list_len; a++)
    {
      cnt += cache_expression (s, &srclist->list.list_val[a], mode);
    }
  return cnt;
}



static char *
get_str_list (str_list * list)
{
  static char buff[2000];
  int a;
  strcpy (buff, "");
  for (a = 0; a < list->str_list_entry.str_list_entry_len; a++)
    {
      if (a)
	strcat (buff, ",");
      strcat (buff, list->str_list_entry.str_list_entry_val[a]);
    }
  return buff;
}

static char *
get_field_list (fh_field_list * list)
{
  static char buff[2000];
  int a;
  strcpy (buff, "");
  for (a = 0; a < list->field_list_entries.field_list_entries_len; a++)
    {
      if (a)
	strcat (buff, ",");
      strcat (buff, expr_as_string_when_possible (list->field_list_entries.field_list_entries_val[a].field.field));
    }
  return buff;
}

static char *
get_var_lit (expr_str * x)
{
  static char buff[256];
  if (x == 0)
    return "";
  switch (x->expr_type)
    {

    case ET_E_V_OR_LIT_NOVALUE:
      return "";
      //case ET_E_V_OR_LIT_VAR: return x->expr_str_u.var->varname;
    case ET_E_V_OR_LIT_INT:
      sprintf (buff, "%d", x->expr_str_u.i);
      return buff;
    case ET_E_V_OR_LIT_STRING:
      return x->expr_str_u.s;
    case ET_E_V_OR_LIT_VAR_AS_STRING:
      return x->expr_str_u.sv;
    case ET_E_V_OR_LIT_IDENT:
      return x->expr_str_u.si;
// case ET_E_V_OR_LIT_NULL: return "";
    default:
      return expr_as_string_when_possible(x);
      A4GL_assertion (1, "Unexpected expression type");

    }
  sprintf (buff, "Bad VAR=%d", x->expr_type);
  return buff;
  return "Blah1";

}
static char *
xml_encode (char *s)
{
  static char buff [1000000];
  int c;
  int a;
  int l;
  int b;
  c = 0;
  if (strchr (s, '&'))
    c++;
  if (strchr (s, '<'))
    c++;
  if (strchr (s, '>'))
    c++;
  if (strchr (s, '"'))
    c++;
  if (strchr (s, '\''))
    c++;
  if (strchr (s, '\n'))
    c++;
  if (strchr (s, '\r'))
    c++;
  if (strchr (s, '\t'))
    c++;
  if (strchr (s, '\\'))
    c++;

  if (c == 0)
    {
      return s;
    }

  l = strlen (s);
  b = 0;
  for (a = 0; a < l; a++)
    {
      if (s[a] == '>')
	{
	  buff[b++] = '&';
	  buff[b++] = 'g';
	  buff[b++] = 't';
	  buff[b++] = ';';
	  continue;
	}

      if (s[a] == '\n') {
		buff[b++]=' '; 
		continue;
	}
      if (s[a] == '\r') {
		continue;
	}
      if (s[a] == '\\' && s[a+1]=='t') { buff[b++]=' '; a++; continue; }

      if (s[a] == '"')
	{
	  buff[b++] = '&';
	  buff[b++] = 'q';
	  buff[b++] = 'u';
	  buff[b++] = 'o';
	  buff[b++] = 't';
	  buff[b++] = ';';
	  continue;
	}
      if (s[a] == '\'')
	{
	  buff[b++] = '&';
	  buff[b++] = 'a';
	  buff[b++] = 'p';
	  buff[b++] = 'o';
	  buff[b++] = 's';
	  buff[b++] = ';';
	  continue;
	}
      if (s[a] == '<')
	{
	  buff[b++] = '&';
	  buff[b++] = 'l';
	  buff[b++] = 't';
	  buff[b++] = ';';
	  continue;
	}

      if (s[a] == '\\' && s[a+1]=='"')
	{
	  buff[b++] = '&';
	  buff[b++] = 'q';
	  buff[b++] = 'u';
	  buff[b++] = 'o';
	  buff[b++] = 't';
	  buff[b++] = ';';
	  a++;
	  continue;
	}

      if (s[a] == '&')
	{
	  buff[b++] = '&';
	  buff[b++] = 'a';
	  buff[b++] = 'm';
	  buff[b++] = 'p';
	  buff[b++] = ';';
	  continue;
	}
      buff[b++] = s[a];
    }

  buff[b] = 0;
  return buff;
}


static char *
get_menu_option (menuoption * mn)
{
  static char buff[1024];
  char smbuff[100];;
  strcpy (buff, "");
  sprintf (smbuff, "KEYS=\"%s\"", xml_encode (get_str_list (mn->keys)));
  strcat (buff, smbuff);
  sprintf (smbuff, " SHORT=\"%s\"", xml_encode (get_var_lit (mn->shortname)));
  strcat (buff, smbuff);
  sprintf (smbuff, " LONG=\"%s\"", xml_encode (get_var_lit (mn->longname)));
  strcat (buff, smbuff);

  return buff;
}

static char *
get_event (event_data * a)
{
  static char buff[1024];
  switch (a->event_type)
    {
    case EVENT_BEFORE_PROMPT:
      return "EVENT_TYPE=\"BEFORE PROMPT\"";
    case EVENT_BEFORE_DIALOG:
      return "EVENT_TYPE=\"BEFORE DIALOG\"";
    case EVENT_AFTER_DIALOG:
      return "EVENT_TYPE=\"AFTER DIALOG\"";
    case EVENT_BEF_ROW:
      return "EVENT_TYPE=\"BEFORE ROW\"";
    case EVENT_AFT_ROW:
      return "EVENT_TYPE=\"AFTTER ROW\"";
    case EVENT_BEFORE_DELETE:
      return "EVENT_TYPE=\"BEFORE DELETE\"";
    case EVENT_AFTER_DELETE:
      return "EVENT_TYPE=\"AFTER DELETE\"";
    case EVENT_BEFORE_INSERT:
      return "EVENT_TYPE=\"BEFORE INSERT\"";
    case EVENT_AFTER_INSERT:
      return "EVENT_TYPE=\"AFTER INSERT\"";
    case EVENT_BEFORE_OPEN_FORM:
      return "EVENT_TYPE=\"BEFORE OPEN FORM\"";
    case EVENT_BEFORE_INSERT_DELETE:
      return "EVENT_TYPE=\"BEFORE INSERT DELETE\"";
    case EVENT_AFTER_INSERT_DELETE:
      return "EVENT_TYPE=\"AFTER INSERT DELETE\"";
    case EVENT_AFTER_INP_CLEAN:
      return "EVENT_TYPE=\"AFTER_INP_CLEAN\"";
    case EVENT_ANYKEY_PRESS:
      return "EVENT_TYPE=\"ANYKEY_PRESS\"";
    case EVENT_AFTER_INPUT:
      return "EVENT_TYPE=\"AFTER INPUT\"";
    case EVENT_BEFORE_INPUT:
      return "EVENT_TYPE=\"BEFORE INPUT\"";

    case EVENT_AFTER_DISPLAY:
      return "EVENT_TYPE=\"AFTER DISPLAY\"";
    case EVENT_BEFORE_DISPLAY:
      return "EVENT_TYPE=\"BEFORE DISPLAY\"";
    case EVENT_AFTER_CONSTRUCT:
      return "EVENT_TYPE=\"AFTER CONSTRUCT\"";
    case EVENT_BEFORE_CONSTRUCT:
      return "EVENT_TYPE=\"BEFORE CONSTRUCT\"";
    case EVENT_BEFORE_MENU:
      return "EVENT_TYPE=\"BEFORE MENU\"";
    case EVENT_AFTER_EVENT:
      return "EVENT_TYPE=\"AFTER EVENT\"";
    case EVENT_BEFORE_CLOSE_FORM:
      return "EVENT_TYPE=\"BEFORE CLOSE_FORM\"";
    case EVENT_BEFORE_ANY:
      return "EVENT_TYPE=\"BEFORE ANY\"";
    case EVENT_AFTER_ANY:
      return "EVENT_TYPE=\"AFTER ANY\"";
    case EVENT_BEFORE_EVENT:
      return "EVENT_TYPE=\"BEFORE EVENT\"";

    case EVENT_ON_ACTION:
      sprintf (buff, "EVENT_TYPE=\"ON ACTION\" DATA=\"%s\"", xml_encode (a->event_data_u.on_action_s->actionName));
      return buff;
    case EVENT_BEFORE:
      sprintf (buff, "EVENT_TYPE=\"BEFORE\" DATA=\"%s\"", xml_encode (get_str_list (a->event_data_u.slist)));
      return buff;

    case EVENT_AFTER:
      sprintf (buff, "EVENT_TYPE=\"AFTER\" DATA=\"%s\"", xml_encode (get_str_list (a->event_data_u.slist)));
      return buff;

    case EVENT_ON:
      sprintf (buff, "EVENT_TYPE=\"ON\" DATA=\"%s\"", xml_encode (get_str_list (a->event_data_u.slist)));
      return buff;

    case EVENT_ON_IDLE:
      return "EVENT_ON_IDLE";
    case EVENT_ON_INTERVAL:
      return "EVENT_ON_INTERVAL";
    case EVENT_ON_TIME:
      return "EVENT_ON_TIME";

    case EVENT_KEY_PRESS:
      sprintf (buff, "EVENT_TYPE=\"KEY_PRESS\" DATA=\"%s\"", xml_encode (get_str_list (a->event_data_u.slist)));
      return buff;
      //case EVENT_ON_ACTION: get_str_list(on_action);
      //case EVENT_ON:                *on;

    case EVENT_BEFORE_FIELD:
      sprintf (buff, "EVENT_TYPE=\"BEFORE_FIELD\" DATA=\"%s\"", xml_encode (get_field_list (a->event_data_u.before_after_field)));
      return buff;

    case EVENT_AFTER_FIELD:
      sprintf (buff, "EVENT_TYPE=\"AFTER_FIELD\" DATA=\"%s\"", xml_encode (get_field_list (a->event_data_u.before_after_field)));
      return buff;

    case EVENT_ON_CHANGE:
      sprintf (buff, "EVENT_TYPE=\"ON_CHANGE\" DATA=\"%s\"", xml_encode (get_field_list (a->event_data_u.before_after_field)));
	return buff;
    case EVENT_MENU_COMMAND:
      sprintf (buff, "EVENT_TYPE=\"MENU_COMMAND\" %s", get_menu_option (a->event_data_u.mnoption));
      return buff;
      //case EVENT_BEFORE:    str_list *before;
      //case EVENT_AFTER:     str_list *after;
    }

  return "Blah2";
}


static int
add_calltree_calls_from_events (char *s, struct on_events *evt_list, int mode)
{
  int a;
  int cnt = 0;
  int xcnt = 0;
  if (evt_list == 0) {
  	if (mode==MODE_TRY) {
  			if (cnt==0 && AlwaysIncludeAllConditionals) {
					return 1;
			}
  	}
    return 0;
  }
  for (a = 0; a < evt_list->event.event_len; a++)
    {

      xcnt = add_calltree_calls (s, evt_list->event.event_val[a]->on_event_commands, MODE_TRY);
      cnt += xcnt;

  			if (xcnt==0 && AlwaysIncludeAllConditionals && mode==MODE_BUY) {
					xcnt++;
					
				}

      if (xcnt)
	{
	  if (mode == MODE_BUY)
	    {
	      print_indent ();
	      fprintf (output, "<EVENT LINE=\"%d\" %s>\n", evt_list->event.event_val[a]->lineno,
		       get_event (&evt_list->event.event_val[a]->evt_data));
	      indent++;
	    }

	  if (mode == MODE_BUY)
	    {
	      print_indent ();
	      fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",
					get_first_lineno(evt_list->event.event_val[a]->on_event_commands ),
					get_last_lineno( evt_list->event.event_val[a]->on_event_commands)
			);
	      indent++;
	    }
	  add_calltree_calls (s, evt_list->event.event_val[a]->on_event_commands, mode);
	  if (mode == MODE_BUY)
	    {

	      indent--;
	      print_indent ();
	      fprintf (output, "</COMMANDS>\n");
	    }

	  if (mode == MODE_BUY)
	    {
	      indent--;
	      print_indent ();
	      fprintf (output, "</EVENT>\n");
	    }
	}
    }

  if (mode==MODE_TRY) {
  		if (cnt==0 && AlwaysIncludeAllConditionals) {
				return 1;
		}
  }
  return cnt;
}


static int
print_whenever (int mode)
{
  int cnt = 0;
  if (whenever_error_func)
    {
      if (mode == MODE_BUY)
	{
	  print_indent ();
	  fprintf (output, "<WHENEVER TYPE=\"ERROR\">");
	  fprintf (output, "<CALLS FUNCTIONNAME='%s' LINE=\"%d\"/>", whenever_error_func, whenever_error_func_line);
	  fprintf (output, "</WHENEVER>\n");
	      addNode (currfunc, whenever_error_func, "",whenever_error_func_line,0, "C");
	}
      cnt++;
    }
  if (whenever_any_error_func)
    {
      if (mode == MODE_BUY)
	{
	  print_indent ();
	  fprintf (output, "<WHENEVER TYPE=\"ANYERROR\">");
	  fprintf (output, "<CALLS FUNCTIONNAME='%s' LINE=\"%d\"/>", whenever_any_error_func, whenever_any_error_func_line);
	  fprintf (output, "</WHENEVER>\n");
	}
      cnt++;
    }
  if (whenever_sql_error_func)
    {
      if (mode == MODE_BUY)
	{
	  print_indent ();
	  fprintf (output, "<WHENEVER TYPE=\"SQLERROR\">");
	  fprintf (output, "<CALLS FUNCTIONNAME='%s' LINE=\"%d\"/>", whenever_sql_error_func, whenever_sql_error_func_line);
	  fprintf (output, "</WHENEVER>\n");
	}
      cnt++;
    }
  return cnt;
}


struct s_symbol_table {
	char *name;
	char *module;
	int line;
	char *type;
	char *operation;
	char *scope;
	char *value;
};

struct s_symbol_table *symbol_table=0;
int nsymbols=0;

static void add_symbol_with_scope(char *name,char*mod,int line,char *type,char *operation,char *scope,char *value) {
int a;
	
	for (a=0;a<nsymbols;a++) {
		if (strcmp(symbol_table[a].name,name)==0 &&
		 	strcmp(symbol_table[a].module,mod)==0 &&
		 	strcmp(symbol_table[a].type,type)==0 &&
		 	strcmp(symbol_table[a].operation,operation)==0 &&
		 	strcmp(symbol_table[a].scope,scope)==0 &&
		 	symbol_table[a].line==line) {
				return;
		}
	}
	nsymbols++;


	symbol_table=realloc(symbol_table, nsymbols*sizeof(symbol_table[0]));
	symbol_table[nsymbols-1].name=strdup(name);
	symbol_table[nsymbols-1].module=strdup(mod);
	symbol_table[nsymbols-1].line=line;
	symbol_table[nsymbols-1].type=strdup(type);
	symbol_table[nsymbols-1].operation=strdup(operation);
	symbol_table[nsymbols-1].scope=strdup(scope);
	if (value) {
	symbol_table[nsymbols-1].value=strdup(value);
	} else {
	symbol_table[nsymbols-1].value="";
	}
}


static void dump_symbols(void) {
int a;
	fprintf(output,"<SYMBOLS>\n");
	for (a=0;a<nsymbols;a++) {
		fprintf(output," <SYMBOL NAME=\"%s\" MODULE=\"%s\" LINE=\"%d\" TYPE=\"%s\" OPERATION=\"%s\" SCOPE=\"%s\" VALUE=\"%s\"/>\n",
			A4GL_strip_quotes(
				symbol_table[a].name),
				symbol_table[a].module,
				symbol_table[a].line,
				symbol_table[a].type,
				symbol_table[a].operation,
				symbol_table[a].scope,
				xml_encode(symbol_table[a].value)
			);
	}
	fprintf(output,"</SYMBOLS>\n");
}


static void add_symbol(char *name,char*mod,int line,char *type,char *operation) {
	add_symbol_with_scope(name,mod,line,type,operation,"G",0);
	//printf("Add symbol : %s (%s:%d) %s %s\n", name,mod,line,type,operation);
}

static void add_variable(expr_str *v, char *module, int lineno, char *operation,expr_str *value) {
char *scope="G";
if (v->expr_type==ET_EXPR_VARIABLE_USAGE) {
	struct variable_usage *u;
	u=v->expr_str_u.expr_variable_usage;
	switch (u->escope) {	
		case E_SCOPE_MODULE:
			scope="M"; break;

		case E_SCOPE_LOCAL:
		case E_SCOPE_REPORT_LOCAL:
			scope="L"; break;
		case E_SCOPE_IMPORTED_GLOBAL:
		case E_SCOPE_EXPORTED_GLOBAL:
			scope="G"; break;
		default: break;
	}

	if (strcmp(operation,"PARAM")==0) {
			operation="ASSIGN";
				scope="L";
		}
	add_symbol_with_scope(expr_as_string_when_possible(v),module,lineno,"VARIABLE",operation,scope,evaluate_expr(value,1));

	if (value && strcmp(operation,"ASSIGN")==0) { 
			add_variable_value(v->expr_str_u.expr_variable_usage, value ); 
	}
}
}

static void add_symbol_assign_single(expr_str *l,char *module,int lineno,expr_str *value) {
	add_variable(l,module,lineno,"ASSIGN",value);
}


static void add_symbol_assign_single_expr(expr_str_list *l,char *module,int lineno,expr_str *value) {
int a;
if (l==0) return;

if (l->list.list_len>2000) {
	printf("Too many values in assignment to add a value for each\n");
	return ;
}
for (a=0;a<l->list.list_len;a++) {
		add_symbol_assign_single(l->list.list_val[a],module,lineno, value);
}

}

static void add_symbol_assign(expr_str_list *l,char *module,int lineno,expr_str_list *values) {
int a;
if (l==0) return;

for (a=0;a<l->list.list_len;a++) {
	if (values!=NULL) {
		// Do we have 1 variable but lots of values ? 
		// This is a concatenation ...
		if (l->list.list_len==1 && values->list.list_len>1) {
		int b;
			char buff[120000]="";
			for (b=0;b< values->list.list_len;b++) {
				strcat(buff, evaluate_expr( values->list.list_val[b],1));
			}
			add_symbol_assign_single(l->list.list_val[a],module,lineno, A4GL_new_literal_string(buff));
		} else {
			add_symbol_assign_single(l->list.list_val[a],module,lineno, values->list.list_val[a]);
		}
	} else {
		add_symbol_assign_single(l->list.list_val[a],module,lineno,NULL);
	}
}

}


static int
calls_something (s_commands * func_commands)
{
int cnt;
 		cnt=add_calltree_calls ("", func_commands, MODE_TRY);
  		if (cnt==0 && AlwaysIncludeAllConditionals) {
				return 1;
		}
  return cnt;
}

struct module_definition cum_clobber;
static void add_clobberings (module_definition * mods, int nmodules) {
int a;
int b;
cum_clobber.clobberings.clobberings_len=0;
cum_clobber.clobberings.clobberings_val=0;
for (a=0;a<nmodules;a++) {
	for (b=0;b< mods[a].clobberings.clobberings_len;b++) {
		cum_clobber.clobberings.clobberings_len++;

		cum_clobber.clobberings.clobberings_val=realloc(
		cum_clobber.clobberings.clobberings_val,
		sizeof(cum_clobber.clobberings.clobberings_val[0])*
		cum_clobber.clobberings.clobberings_len);

		cum_clobber.clobberings.clobberings_val[

		cum_clobber.clobberings.clobberings_len-1].newval=
			mods[a].clobberings.clobberings_val[b].newval;

		cum_clobber.clobberings.clobberings_val[
		cum_clobber.clobberings.clobberings_len-1].important=
			mods[a].clobberings.clobberings_val[b].important;
	}
}
}



static char *
get_orig_from_clobber (struct module_definition *mod, char *s)
{
  int a;

  if (mod == NULL)
    return s;

  for (a = 0; a < mod->clobberings.clobberings_len; a++)
    {
      if (strcmp (mod->clobberings.clobberings_val[a].newval, s) == 0)
        {
          return mod->clobberings.clobberings_val[a].important;
        }
    }

  return s;
}


static char *calltree_get_ident(struct expr_str *ptr) {
 static char buff[2000];
  if (ptr->expr_type == ET_EXPR_IDENTIFIER)
    {
      sprintf (buff, "%s", get_orig_from_clobber (&cum_clobber, ptr->expr_str_u.expr_string));
      return strdup (buff);
    }

  if (ptr->expr_type == ET_EXPR_VARIABLE_IDENTIFIER)    // a _VARIABLE
    {
      return
        strdup (expr_as_string_when_possible (ptr->expr_str_u.expr_expr));
    }

  if (ptr->expr_type == ET_EXPR_VARIABLE_USAGE)
    {
      return expr_as_string_when_possible(ptr);
    }


  A4GL_assertion (1,
                  "lint_get_ident_as_string not implemented for this expression type yet");
  return 0;

}



static char *A4GL_strip_quotes_1(char *s) {
static char ptr[200000];
	int a;
int b=0;
if (s[0]=='\"') {
	s=A4GL_strip_quotes(s);
	for (a=0;a<strlen(s);a++) {
	
		if (s[a]=='\t') {
			ptr[b++]=' '; continue;
		}

		if (s[a]=='\\' && s[a+1]=='t') {
			a++;
			ptr[b++]=' '; continue;
		}
	
		if (s[a]=='\\' && s[a+1]=='\\' && s[a+2]=='t') {
			a+=2;
			ptr[b++]=' '; continue;
		}
		if (s[a]=='\\' && s[a+1]=='\\' && s[a+2]=='n') {
			a+=2;
			ptr[b++]=' '; continue;
		}
		if (s[a]=='\\' && s[a+1]=='\\' && s[a+2]=='"') {
			a+=2;
			ptr[b++]=s[a];
		} else {
			ptr[b++]=s[a];
		}
	}
}  else {
	for (a=0;a<strlen(s);a++) {
	
		if (s[a]=='\t') {
			ptr[b++]=' '; continue;
		}
	
		if (s[a]=='\\' && s[a+1]=='t') {
			a++;
			ptr[b++]=' '; continue;
		}
		if (s[a]=='\\' && s[a+1]=='n') {
			a++;
			ptr[b++]=' '; continue;
		}
		if (s[a]=='\\' && s[a+1]=='"') {
			a++;
			ptr[b++]=s[a];
		} else {
			ptr[b++]=s[a];
		}
	}
}
ptr[b]=0;
return ptr;
}

static int
add_calltree_calls (char *s, s_commands * func_commands, int mode)
{
  int a;
  int call_cnt = 0;




  if (func_commands == 0)
    return 0;

  call_cnt += print_whenever (MODE_TRY);



  for (a = 0; a < func_commands->cmds.cmds_len; a++)
    {
	 //printf("Process cmd : %d\n", a);
		call_cnt+=process_cmd(func_commands->cmds.cmds_val[a],mode);
    }

  return call_cnt;
}


static int process_cmd(struct command *cmd,int mode) {
	int call_cnt=0;
  struct on_events *evt_list;
  struct struct_display_cmd *disp;
	last_mod=cmd->module;
	last_line=cmd->lineno;


      switch (cmd->cmd_data.type)
	{



	case E_CMD_INIT_CMD:
		add_symbol_assign_single_expr(cmd->cmd_data.command_data_u.init_cmd.varlist,last_mod,last_line,A4GL_new_expr_simple(ET_EXPR_NULL));
		break;


	case E_CMD_DELETE_CMD:
		calltree_map_insert_delete_update("DELETE", cmd->cmd_data.command_data_u.delete_cmd.table, 
				cmd->cmd_data.command_data_u.delete_cmd.where_clause, last_mod,last_line);
		break;

	case E_CMD_SQL_CMD:
			calltree_map_sql(cmd->cmd_data.command_data_u.sql_cmd.sql,last_mod,last_line);
		break;

	case E_CMD_INSERT_CMD:
		calltree_map_insert_delete_update("INSERT", cmd->cmd_data.command_data_u.insert_cmd.table, NULL, last_mod,last_line);

		if (cmd->cmd_data.command_data_u.insert_cmd.subselect) {
			calltree_map_select_stmt("INSERT", cmd->cmd_data.command_data_u.insert_cmd.subselect, last_mod,last_line);
		}
		if (cmd->cmd_data.command_data_u.insert_cmd.value_list) {
			calltree_map_value_list("INSERT", cmd->cmd_data.command_data_u.insert_cmd.value_list, last_mod,last_line);
		}
		if (mode==MODE_BUY) {
		}
		break;

	case E_CMD_UPDATE_CMD:
		calltree_map_insert_delete_update("UPDATE", cmd->cmd_data.command_data_u.update_cmd.table, 
				cmd->cmd_data.command_data_u.update_cmd.where_clause, last_mod,last_line);
		calltree_map_value_list("UPDATE", cmd->cmd_data.command_data_u.update_cmd.value_list, last_mod,last_line);
		if (mode==MODE_BUY) {
		}
		break;

	case E_CMD_SELECT_CMD:
		if (mode==MODE_BUY || 1) {
		 	calltree_map_select_stmt("SELECT", cmd->cmd_data.command_data_u.select_cmd.sql, last_mod,last_line);
		}
		add_symbol_assign(cmd->cmd_data.command_data_u.select_cmd.sql->into,last_mod,last_line,NULL);
		if (A4GL_isyes(acl_getenv("LOGJOINS"))) {
			preprocess_sql_statement ( cmd->cmd_data.command_data_u.select_cmd.sql);
		}
		break;

	case E_CMD_EXECUTE_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.execute_cmd.sql_stmtid),last_mod,last_line,"STMT","EXECUTE");
		//add_symbol_use(cmd->cmd_data.command_data_u.execute_cmd.outbind,last_mod,last_line,NULL);

		add_symbol_assign(cmd->cmd_data.command_data_u.execute_cmd.outbind,last_mod,last_line,NULL);
	  	call_cnt += cache_expression_list ("", cmd->cmd_data.command_data_u.execute_cmd.inbind, mode);

	break;
		break;

	case E_CMD_LOCATE_CMD:
		add_symbol_assign(cmd->cmd_data.command_data_u.locate_cmd.variables,last_mod,last_line,NULL);
		break;




	case E_CMD_WHENEVER_CMD:
	  set_whenever (cmd);
	  call_cnt += print_whenever (mode);
	  break;


	case E_CMD_FOR_CMD:

	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.for_cmd.start, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.for_cmd.end, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.for_cmd.step, mode);

	  if (calls_something (cmd->cmd_data.command_data_u.for_cmd.for_commands))
	    {
	      call_cnt++;
	      if (mode == MODE_BUY)
		{
		  print_indent ();
		  fprintf (output, "<FOR START=\"%s\"",
			   xml_encode (expr_as_string_when_possible
				       (cmd->cmd_data.command_data_u.for_cmd.start)));
		  fprintf (output, " END=\"%s\"",
			   xml_encode (expr_as_string_when_possible
				       (cmd->cmd_data.command_data_u.for_cmd.end)));
		  fprintf (output, " STEP=\"%s\" LINE=\"%d\">\n",
			   xml_encode (expr_as_string_when_possible
				       (cmd->cmd_data.command_data_u.for_cmd.step)),
			   cmd->lineno);
		  indent++;
		  print_indent ();
		  fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",  
		  			get_first_lineno(cmd->cmd_data.command_data_u.for_cmd.for_commands),
		  			get_last_lineno(cmd->cmd_data.command_data_u.for_cmd.for_commands)
		  );
		  indent++;
		  add_calltree_calls ("", cmd->cmd_data.command_data_u.for_cmd.for_commands, mode);
		  indent--;
		  print_indent ();
		  fprintf (output, "</COMMANDS>\n");
		  indent--;
		  print_indent ();
		  fprintf (output, "</FOR>\n");
		}
	    }
	  break;

	case E_CMD_IF_CMD:
	  {
	    struct struct_if_cmd *ifcmd;
	    int xcnt;
	    int printed_if = 0;

	    inc_inIf();

	    ifcmd = &cmd->cmd_data.command_data_u.if_cmd;

	    for (xcnt = 0; xcnt < ifcmd->truths.conditions.conditions_len; xcnt++)
	      {
		call_cnt += cache_expression ("", &ifcmd->truths.conditions.conditions_val[xcnt].test_expr, mode);

	      }

	//printf("%d\n",call_cnt);


	    for (xcnt = 0; xcnt < cmd->cmd_data.command_data_u.if_cmd.truths.conditions.conditions_len;
		 xcnt++)
	      {

		ifBlocks[inIf]++;

		if (calls_something
		    (cmd->cmd_data.command_data_u.if_cmd.truths.conditions.conditions_val[xcnt].
		     whentrue))
		  {
		    call_cnt++;
		    if (mode == MODE_BUY)
		      {
			if (!printed_if)
			  {
			    printed_if++;
			    print_indent ();
			    fprintf (output, "<IF CONDITION=\"%s\" LINE=\"%d\">\n",
				     xml_encode (expr_as_string_when_possible
						 (ifcmd->truths.conditions.conditions_val[xcnt].test_expr)),
						 cmd->lineno
					  	);
			  }
			indent++;
			print_indent ();
			fprintf (output, "<ONTRUE>\n");
			indent++;
			print_indent ();
			fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n", 
									get_first_lineno(cmd->cmd_data.command_data_u.if_cmd.truths.conditions.  conditions_val[xcnt].whentrue),
									get_last_lineno(cmd->cmd_data.command_data_u.if_cmd.truths.conditions.  conditions_val[xcnt].whentrue)
									 	);
			indent++;
			add_calltree_calls ("",
					    cmd->cmd_data.command_data_u.if_cmd.truths.conditions.
					    conditions_val[xcnt].whentrue, mode);
			indent--;
			print_indent ();
			fprintf (output, "</COMMANDS>\n");
			indent--;
			print_indent ();
			fprintf (output, "</ONTRUE>\n");
			indent--;
		      }
		  }
	      }

	    if (cmd->cmd_data.command_data_u.if_cmd.whenfalse)
	      {
		ifBlocks[inIf]++;
		if (calls_something (cmd->cmd_data.command_data_u.if_cmd.whenfalse))
		  {
		    call_cnt++;
		    if (mode == MODE_BUY)
		      {
			if (!printed_if)
			  {
			    printed_if++;
			    print_indent ();
			    fprintf (output, "<IF CONDITION=\"%s\" LINE=\"%d\">\n",
				     xml_encode (expr_as_string_when_possible
						 (ifcmd->truths.conditions.conditions_val[0].test_expr)),
				     cmd->cmd_data.command_data_u.if_cmd.lineno);
			  }
			indent++;
			print_indent ();
			fprintf (output, "<ONFALSE LINE=\"%d\">\n",
				 cmd->cmd_data.command_data_u.if_cmd.else_lineno);
			indent++;
			print_indent ();
			fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n", 
						get_first_lineno(cmd->cmd_data.command_data_u.if_cmd.whenfalse),
						get_last_lineno(cmd->cmd_data.command_data_u.if_cmd.whenfalse));
			indent++;
			add_calltree_calls ("", cmd->cmd_data.command_data_u.if_cmd.whenfalse, mode);
			indent--;
			print_indent ();
			fprintf (output, "</COMMANDS>\n");
			indent--;
			print_indent ();
			fprintf (output, "</ONFALSE>\n");
			indent--;
		      }

		  }

	      }
	    if (printed_if)
	      {
		print_indent ();
		fprintf (output, "</IF>\n");
	      }
		dec_inIf();
	  }
	  break;

	case E_CMD_FOREACH_CMD:
		add_symbol_assign(cmd->cmd_data.command_data_u.foreach_cmd.outputvals,last_mod,last_line,NULL);
		call_cnt+=cache_expression_list( "", cmd->cmd_data.command_data_u.foreach_cmd.inputvals, mode);
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.foreach_cmd.cursorname),last_mod,last_line,"CURSOR","FETCH");

	  if (calls_something (cmd->cmd_data.command_data_u.foreach_cmd.foreach_commands))
	    {
	      call_cnt++;
	      if (mode == MODE_BUY)
		{
		  print_indent ();
		  fprintf (output, "<FOREACH LINE=\"%d\">\n", cmd->lineno);
		  indent++;
		  print_indent ();
		  fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n", 
		  				get_first_lineno(cmd->cmd_data.command_data_u.foreach_cmd.foreach_commands),
		  				get_last_lineno(cmd->cmd_data.command_data_u.foreach_cmd.foreach_commands));
		  indent++;
		  add_calltree_calls ("", cmd->cmd_data.command_data_u.foreach_cmd.foreach_commands,
				      mode);
		  indent--;
		  print_indent ();
		  fprintf (output, "</COMMANDS>\n");
		  indent--;
		  print_indent ();
		  fprintf (output, "</FOREACH>\n");
		}
	    }
	  break;

	case E_CMD_WHILE_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.while_cmd.while_expr, mode);
	  if (calls_something (cmd->cmd_data.command_data_u.while_cmd.while_commands))
	    {
	      call_cnt++;

	      if (mode == MODE_BUY)
		{
		  print_indent ();
		  fprintf (output, "<WHILE CONDITION=\"%s\" LINE=\"%d\">\n",
			   xml_encode (expr_as_string_when_possible
				       (cmd->cmd_data.command_data_u.while_cmd.while_expr)),
			   cmd->lineno);
		  indent++;
		  print_indent ();
		  fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",
		  			get_first_lineno(cmd->cmd_data.command_data_u.while_cmd.while_commands),
						get_last_lineno(cmd->cmd_data.command_data_u.while_cmd.while_commands)
		  		);
		  indent++;
		  add_calltree_calls ("", cmd->cmd_data.command_data_u.while_cmd.while_commands, mode);
		  indent--;
		  print_indent ();
		  fprintf (output, "</COMMANDS>\n");
		  indent--;
		  print_indent ();
		  fprintf (output, "</WHILE>\n");
		}
	    }
	  break;


	case E_CMD_CASE_CMD:
	  {
	    struct struct_case_cmd *casecmd;
	    int printed_case = 0;
	    int xcnt;

   	    inc_inIf();

	    casecmd = &cmd->cmd_data.command_data_u.case_cmd;


	    cache_expression ("", &casecmd->case_expr, mode);

	    for (xcnt = 0; xcnt < cmd->cmd_data.command_data_u.case_cmd.whens->whens.whens_len; xcnt++)
	      {

		call_cnt += cache_expression ("", &casecmd->whens->whens.whens_val[xcnt]->when_expr, mode);


		if (calls_something
		    (cmd->cmd_data.command_data_u.case_cmd.whens->whens.whens_val[xcnt]->when_commands))
		  {
			ifBlocks[inIf]++;
		    call_cnt++;
		    if (mode == MODE_BUY)
		      {
			if (!printed_case)
			  {
			    if (casecmd->case_expr)
			      {
				print_indent ();
				fprintf (output, "<CASE TESTAGAINST=\"%s\" LINE=\"%d\">\n",
					 xml_encode (expr_as_string_when_possible (casecmd->case_expr)),
					 cmd->lineno);
				printed_case = 1;
				indent++;
			      }
			    else
			      {
				print_indent ();
				fprintf (output, "<CASE  LINE=\"%d\">\n", cmd->lineno);
				printed_case = 1;
				indent++;
			      }
			  }
			print_indent ();
			fprintf (output, "<WHEN CONDITION=\"%s\" LINE=\"%d\">\n",
				 xml_encode (expr_as_string_when_possible (casecmd->whens->whens.whens_val[xcnt]->when_expr)),
				 casecmd->whens->whens.whens_val[xcnt]->lineno);
			indent++;
			print_indent ();
			fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",
					get_first_lineno(cmd->cmd_data.command_data_u.case_cmd.whens->whens.
					                   whens_val[xcnt]->when_commands),
					get_last_lineno(cmd->cmd_data.command_data_u.case_cmd.whens->whens.
					                   whens_val[xcnt]->when_commands)
				);
			indent++;
			add_calltree_calls ("",
					    cmd->cmd_data.command_data_u.case_cmd.whens->whens.
					    whens_val[xcnt]->when_commands, mode);
			indent--;
			print_indent ();
			fprintf (output, "</COMMANDS>\n");
			indent--;
			print_indent ();
			fprintf (output, "</WHEN>\n");
		      }
		  }
	      }


	    if (cmd->cmd_data.command_data_u.case_cmd.otherwise)
	      {
		if (calls_something (cmd->cmd_data.command_data_u.case_cmd.otherwise))
		  {
			ifBlocks[inIf]++;
		    call_cnt++;
		    if (mode == MODE_BUY)
		      {
			if (!printed_case)
			  {
			    print_indent ();
			    fprintf (output, "<CASE LINE=\"%d\">\n", cmd->lineno);
			    printed_case = 1;
			    indent++;
			  }
			indent++;
			print_indent ();
			fprintf (output, "<OTHERWISE>\n");
			print_indent ();
			fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",
					get_first_lineno(cmd->cmd_data.command_data_u.case_cmd.otherwise ),
					get_last_lineno( cmd->cmd_data.command_data_u.case_cmd.otherwise));
			indent++;
			add_calltree_calls ("", cmd->cmd_data.command_data_u.case_cmd.otherwise, mode);
			indent--;
			print_indent ();
			fprintf (output, "</COMMANDS>\n");
			print_indent ();
			fprintf (output, "</OTHERWISE>\n");
			indent--;
		      }
		  }
	      }

	    if (printed_case && mode == MODE_BUY)
	      {
		indent--;
		print_indent ();
		fprintf (output, "</CASE>\n");
	      }
		dec_inIf();
	  }

	  break;


	case E_CMD_LOAD_CMD:
	{
	  if (mode == MODE_BUY)
	    {
			add_file_operation("LOAD",cmd->cmd_data.command_data_u.load_cmd.filename,last_mod,cmd->lineno);
	    }
	}
	break;

	case E_CMD_UNLOAD_CMD:
	{
	  if (mode == MODE_BUY)
	    {
			add_file_operation("UNLOAD",cmd->cmd_data.command_data_u.unload_cmd.filename,last_mod,cmd->lineno);
	    }
	}
	break;


	case E_CMD_START_CMD:

	  if (mode == MODE_BUY)
	    {
	      print_indent ();
	      fprintf (output, "<START REPORT=\"%s\" LINE=\"%d\"/>\n",
		       cmd->cmd_data.command_data_u.start_cmd.repname,
		       cmd->lineno);
	      addNode (currfunc, cmd->cmd_data.command_data_u.start_cmd.repname, "",cmd->lineno,0,"C");
	      if (cmd->cmd_data.command_data_u.start_cmd.sc_c->towhat=='F') {
			add_file_operation("STARTREP",cmd->cmd_data.command_data_u.start_cmd.sc_c->s1,last_mod,cmd->lineno);
//static void add_file_operation(char *operation, expr_str *file, char *module,int line) {
		}

	    }
	  call_cnt++;
	  break;

	case E_CMD_FINISH_CMD:
	  if (mode == MODE_BUY)
	    {
	      print_indent ();
	      fprintf (output, "<FINISH REPORT=\"%s\" LINE=\"%d\"/>\n",
		       cmd->cmd_data.command_data_u.finish_cmd.repname,
		       cmd->lineno);
	      addNode (currfunc, cmd->cmd_data.command_data_u.finish_cmd.repname, "",cmd->lineno,0,"C");
	    }
	  call_cnt++;
	  break;

	case E_CMD_OUTPUT_CMD:
	  cache_expression_list ("", cmd->cmd_data.command_data_u.output_cmd.expressions, mode);
	  if (mode == MODE_BUY)
	    {
	      print_indent ();
	      fprintf (output, "<OUTPUT REPORT=\"%s\" LINE=\"%d\"/>\n",
		       cmd->cmd_data.command_data_u.output_cmd.repname,
		       cmd->lineno);
	      addNode (currfunc, cmd->cmd_data.command_data_u.output_cmd.repname, "",cmd->lineno,0, "C");
	    }
	  call_cnt++;
	  break;


	case E_CMD_MENU_CMD:
	  {
	    int does_call;
	    evt_list = cmd->cmd_data.command_data_u.menu_cmd.events;
	    does_call = add_calltree_calls_from_events ("", evt_list, MODE_TRY);
	    if (does_call)
	      {
		call_cnt += does_call;
		if (mode == MODE_BUY)
		  {
		    print_indent ();
		    fprintf (output, "<MENU LINE=\"%d\">\n", cmd->lineno);
		    indent++;
		    print_indent ();
		    fprintf (output, "<EVENTS>\n");
		    indent++;
		    add_calltree_calls_from_events ("", evt_list, mode);
		    indent--;
		    print_indent ();
		    fprintf (output, "</EVENTS>\n");
		    indent--;
		    print_indent ();
		    fprintf (output, "</MENU>\n");
		  }
	      }
	  }
	  break;

	case E_CMD_PROMPT_CMD:
	  {
	    int does_call;
	    evt_list = cmd->cmd_data.command_data_u.prompt_cmd.events;


	add_symbol_assign_single(cmd->cmd_data.command_data_u.prompt_cmd.promptvar,last_mod,last_line,NULL);

	    does_call = add_calltree_calls_from_events ("", evt_list, MODE_TRY);
	    if (does_call)
	      {
		call_cnt += does_call;
		if (mode == MODE_BUY)
		  {
		    print_indent ();
		    fprintf (output, "<PROMPT LINE=\"%d\">\n", cmd->lineno);
		    indent++;
		    print_indent ();
		    fprintf (output, "<EVENTS>\n");
		    indent++;
		    add_calltree_calls_from_events ("", evt_list, mode);
		    indent--;
		    print_indent ();
		    fprintf (output, "</EVENTS>\n");
		    indent--;
		    print_indent ();
		    fprintf (output, "</PROMPT>\n");
		  }
	      }
	  }
	  break;

	case E_CMD_DISPLAY_ARRAY_CMD:
	  {
	    int does_call;
	    evt_list = cmd->cmd_data.command_data_u.display_array_cmd.events;
	    does_call = add_calltree_calls_from_events ("", evt_list, MODE_TRY);
		add_variable(cmd->cmd_data.command_data_u.display_array_cmd.arrayname,
					cmd->module,
								cmd->lineno,"USE",NULL);
	    if (does_call)
	      {
		call_cnt += does_call;
		if (mode == MODE_BUY)
		  {
		    print_indent ();
		    fprintf (output, "<DISPLAYARRAY LINE=\"%d\">\n", cmd->lineno);
		    indent++;
		    print_indent ();
		    fprintf (output, "<EVENTS>\n");
		    indent++;
		    add_calltree_calls_from_events ("", evt_list, mode);
		    indent--;
		    print_indent ();
		    fprintf (output, "</EVENTS>\n");
		    indent--;
		    print_indent ();
		    fprintf (output, "</DISPLAYARRAY>\n");
		  }
	      }
	  }
	  break;


	case E_CMD_INPUT_CMD:
	  {
	    int does_call;
		add_symbol_assign(cmd->cmd_data.command_data_u.input_cmd.variables,last_mod,last_line,NULL);
	    evt_list = cmd->cmd_data.command_data_u.input_cmd.events;
	    does_call = add_calltree_calls_from_events ("", evt_list, MODE_TRY);
	    if (does_call)
	      {
		call_cnt += does_call;
		if (mode == MODE_BUY)
		  {
		    print_indent ();
		    fprintf (output, "<INPUT LINE=\"%d\">\n", cmd->lineno);
		    indent++;
		    print_indent ();
		    fprintf (output, "<EVENTS>\n");
		    indent++;
		    add_calltree_calls_from_events ("", evt_list, mode);
		    indent--;
		    print_indent ();
		    fprintf (output, "</EVENTS>\n");
		    indent--;
		    print_indent ();
		    fprintf (output, "</INPUT>\n");
		  }
	      }
	  }
	  break;

	case E_CMD_INPUT_ARRAY_CMD:
	  {
	    int does_call;
		add_symbol_assign_single(cmd->cmd_data.command_data_u.input_array_cmd.arrayname,last_mod,last_line,NULL);
	    evt_list = cmd->cmd_data.command_data_u.input_array_cmd.events;
	    does_call = add_calltree_calls_from_events ("", evt_list, MODE_TRY);
	    if (does_call)
	      {
		call_cnt += does_call;
		if (mode == MODE_BUY)
		  {
		    print_indent ();
		    fprintf (output, "<INPUTARRAY LINE=\"%d\">\n", cmd->lineno);
		    indent++;
		    print_indent ();
		    fprintf (output, "<EVENTS>\n");
		    indent++;
		    add_calltree_calls_from_events ("", evt_list, mode);
		    indent--;
		    print_indent ();
		    fprintf (output, "</EVENTS>\n");
		    indent--;
		    print_indent ();
		    fprintf (output, "</INPUTARRAY>\n");
		  }
	      }
	  }
	  break;

	case E_CMD_CONSTRUCT_CMD:
	  {
	    int does_call;
		add_symbol_assign_single(cmd->cmd_data.command_data_u.construct_cmd.constr_var,last_mod,last_line,NULL);
	    evt_list = cmd->cmd_data.command_data_u.construct_cmd.events;
	    does_call = add_calltree_calls_from_events ("", evt_list, MODE_TRY);
	    if (does_call)
	      {
		call_cnt += does_call;
		if (mode == MODE_BUY)
		  {
		    print_indent ();
		    fprintf (output, "<CONSTRUCT LINE=\"%d\">\n", cmd->lineno);
		    indent++;
		    print_indent ();
		    fprintf (output, "<EVENTS>\n");
		    indent++;
		    add_calltree_calls_from_events ("", evt_list, mode);
		    indent--;
		    print_indent ();
		    fprintf (output, "</EVENTS>\n");
		    indent--;
		    print_indent ();
		    fprintf (output, "</CONSTRUCT>\n");
		  }
	      }
	    break;
	  }





	case E_CMD_DISPLAY_CMD:
	  disp = &cmd->cmd_data.command_data_u.display_cmd;
	  call_cnt += cache_expression_list ("", disp->exprs, mode);
	  if (disp->where->dttype == DT_DISPLAY_TYPE_AT)
	    {
	      call_cnt += cache_expression ("", &disp->where->dt_display_u.x_y.x, mode);
	      call_cnt += cache_expression ("", &disp->where->dt_display_u.x_y.y, mode);
	    }
	  break;


	case E_CMD_ERROR_CMD:
	  call_cnt +=
	    cache_expression_list ("", cmd->cmd_data.command_data_u.error_cmd.expr_list, mode);
	  break;

	case E_CMD_SLEEP_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.sleep_cmd.sleep_expr, mode);
	  break;

	case E_CMD_MOVE_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.move_cmd.windowname),last_mod,last_line,"WINDOW","MOVE");
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.move_cmd.x, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.move_cmd.y, mode);
	  break;

	case E_CMD_SHOW_MENU_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.show_menu_cmd.mn_file, mode);
	  break;

	case E_CMD_RUN_CMD:
	{
		char buff[20000];
		sprintf(buff,"<RUN_STMT CMD=\"%s\" MODULE=\"%s\" LINE=\"%d\"/>\n",xml_encode(evaluate_expr(cmd->cmd_data.command_data_u.run_cmd.run_string,1)),last_mod,last_line);
		calltree_addmap(MAPSET_RUN ,buff,last_mod,last_line);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.run_cmd.run_string, mode);
	}
	  break;

	case E_CMD_RUN_WAITING_FOR_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.run_waiting_for_cmd.run_string, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.run_waiting_for_cmd.sleep, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.run_waiting_for_cmd.msg_repeat_every, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.run_waiting_for_cmd.msg_text, mode);
	  break;

	case E_CMD_SCROLL_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.scroll_cmd.val, mode);
	  break;
	case E_CMD_EXIT_PROG_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.exit_prog_cmd.exit_val, mode);
	  break;

	case E_CMD_RETURN_CMD:
	  call_cnt += cache_expression_list ("", cmd->cmd_data.command_data_u.return_cmd.retvals, mode);
	  break;

	case E_CMD_MESSAGE_CMD:
	  call_cnt +=
	    cache_expression_list ("", cmd->cmd_data.command_data_u.message_cmd.message_expr, mode);
	  break;


	case E_CMD_CALL_CMD:
		add_symbol_assign(cmd->cmd_data.command_data_u.call_cmd.returning,last_mod,last_line,NULL);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.call_cmd.fcall, mode);
	  break;

	case E_CMD_PDF_CALL_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.pdf_call_cmd.fcall, mode);
	  break;

	case E_CMD_LET_CMD:
		add_symbol_assign(cmd->cmd_data.command_data_u.let_cmd.vars,last_mod,last_line, cmd->cmd_data.command_data_u.let_cmd.vals );
	  call_cnt += cache_expression_list ("", cmd->cmd_data.command_data_u.let_cmd.vals, mode);
	  break;

	case E_CMD_DECLARE_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.declare_cmd.cursorname),last_mod,last_line,"CURSOR","DECLARE");
		if (cmd->cmd_data.command_data_u.declare_cmd.declare_dets->ident) {
			add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.declare_cmd.declare_dets->ident),last_mod,last_line,"STMT","DECLARE");
		}

		if (cmd->cmd_data.command_data_u.declare_cmd.declare_dets->select) {
			if (mode==MODE_BUY || 1) {
		 		calltree_map_select_stmt("SELECT", cmd->cmd_data.command_data_u.declare_cmd.declare_dets->select, last_mod,last_line);
			}
			add_symbol_assign(cmd->cmd_data.command_data_u.declare_cmd.declare_dets->select->into,last_mod,last_line,NULL);
			if (A4GL_isyes(acl_getenv("LOGJOINS"))) {
				preprocess_sql_statement (cmd->cmd_data.command_data_u.declare_cmd.declare_dets->select);
			}
		}

		if (cmd->cmd_data.command_data_u.declare_cmd.declare_dets->insert_cmd) {
			calltree_map_insert_delete_update("INSERT", cmd->cmd_data.command_data_u.declare_cmd.declare_dets->insert_cmd->table, NULL, last_mod,last_line);

			if (cmd->cmd_data.command_data_u.insert_cmd.subselect) {
				calltree_map_select_stmt("INSERT", cmd->cmd_data.command_data_u.declare_cmd.declare_dets->insert_cmd->subselect, last_mod,last_line);
			}
			if (cmd->cmd_data.command_data_u.insert_cmd.value_list) {
				calltree_map_value_list("INSERT", cmd->cmd_data.command_data_u.declare_cmd.declare_dets->insert_cmd->value_list, last_mod,last_line);
		}
		}

		break;


	case E_CMD_PREPARE_CMD:
		{
		struct command *c;
		char *str;
		str=evaluate_expr(cmd->cmd_data.command_data_u.prepare_cmd.sql,0);
		str=A4GL_strip_quotes_1(str);
		printf("Add prepare : %s=%s\n", (calltree_get_ident(cmd->cmd_data.command_data_u.prepare_cmd.stmtid)), str);
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.prepare_cmd.stmtid),last_mod,last_line,"STMT","PREPARE");
	  	call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.prepare_cmd.sql, mode);



		c=processSQL(str);
		printf("Prepare - c=%p\n",c);

		if (c!=0) {
			c->lineno=last_line;
			c->module=last_mod;
			process_cmd(c,mode);
			guess_sql_stmt(&cmd->cmd_data.command_data_u.prepare_cmd, last_mod,last_line,1);
			if (c->cmd_data.type==E_CMD_SELECT_CMD ) {
				if (A4GL_isyes(acl_getenv("LOGJOINS"))) {
					preprocess_sql_statement ( c->cmd_data.command_data_u.select_cmd.sql);
				}
			}
		} else {
			printf("Prepare @ %s:%d not handled\n", last_mod,last_line);
			guess_sql_stmt(&cmd->cmd_data.command_data_u.prepare_cmd, last_mod,last_line,0);
		}
		
		break;

	case E_CMD_OPEN_CURSOR_CMD:
		call_cnt+=cache_expression_list( "", cmd->cmd_data.command_data_u.open_cursor_cmd.using_bind, mode);
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.open_cursor_cmd.cursorname),last_mod,last_line,"CURSOR","OPEN");
		break;

	case E_CMD_CLOSE_CMD:
		if (cmd->cmd_data.command_data_u.close_cmd.cl_type==E_CT_WINDOW) {
			add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.close_cmd.ident),last_mod,last_line,"WINDOW","CLOSE");
		} 

		if (cmd->cmd_data.command_data_u.close_cmd.cl_type==E_CT_FORM) {
			add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.close_cmd.ident),last_mod,last_line,"FORM","CLOSE");
		} 
		break;

	case E_CMD_CLOSE_SQL_CMD:
		if (cmd->cmd_data.command_data_u.close_sql_cmd.cl_type==E_CT_CURS_OR_PREP) {
			// We dont know which - so add them both and let the calltree viewer sort it out..
			add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.close_sql_cmd.ident),last_mod,last_line,"CURSOR","CLOSE");
			//add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.close_sql_cmd.ident),last_mod,last_line,"STMT","CLOSE");
		}
		break;

	case E_CMD_CURRENT_WIN_CMD:
		if (cmd->cmd_data.command_data_u.current_win_cmd.windowname) {
			add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.current_win_cmd.windowname),last_mod,last_line,"WINDOW","CURRENT");
		}
		break;

	case E_CMD_HIDE_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.hide_cmd.windowname),last_mod,last_line,"WINDOW","HIDE");
		break;
	case E_CMD_SHOW_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.show_cmd.windowname),last_mod,last_line,"WINDOW","SHOW");
		break;

	case E_CMD_OPEN_WINDOW_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.open_window_cmd.windowname),last_mod,last_line,"WINDOW","OPEN");
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.open_window_cmd.x, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.open_window_cmd.y, mode);
	  if (cmd->cmd_data.command_data_u.open_window_cmd.wt.wintype == EWT_FORM)
	    {
		char buff[20000];
	      call_cnt += cache_expression ("",
					    &cmd->cmd_data.command_data_u.open_window_cmd.wt.
					    windowtype_u.formfilename, mode);
		sprintf(buff,"<FORM SOURCE=\"%s\" MODULE=\"%s\" LINE=\"%d\">%s</FORM>\n",xml_encode(evaluate_expr(
			cmd->cmd_data.command_data_u.open_window_cmd.wt.windowtype_u.formfilename,1)) ,
			last_mod,last_line,get_form_display(evaluate_expr(
                        cmd->cmd_data.command_data_u.open_window_cmd.wt.windowtype_u.formfilename,1)));

		calltree_addmap(MAPSET_FORM ,buff,last_mod,last_line);
	    }
	  break;

	case E_CMD_PUT_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.put_cmd.cursorname),last_mod,last_line,"CURSOR","PUT");
		break;

	case E_CMD_FREE_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.free_cmd.cursorname),last_mod,last_line,"CURSOR","FREE");
		break;
	case E_CMD_FLUSH_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.flush_cmd.cursorname),last_mod,last_line,"CURSOR","FLUSH");
		break;

	case E_CMD_FETCH_CMD:
	add_symbol_assign(cmd->cmd_data.command_data_u.fetch_cmd.outbind,last_mod,last_line,NULL);

	add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.fetch_cmd.fetch->cursorname),last_mod,last_line,"CURSOR","FETCH");
	  if (cmd->cmd_data.command_data_u.fetch_cmd.fetch)
	    {
	      if (cmd->cmd_data.command_data_u.fetch_cmd.fetch->fp)
		{
		  call_cnt +=
		    cache_expression ("", &cmd->cmd_data.command_data_u.fetch_cmd.fetch->fp->fetch_expr,
				      mode);
		}
	    }
	  break;

	case E_CMD_DISPLAY_FORM_CMD:
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.display_form_cmd.formname),last_mod,last_line,"FORM","DISPLAY");
	break;
	case E_CMD_OPEN_FORM_CMD:
		{
		char buff[20000];
		add_symbol(calltree_get_ident(cmd->cmd_data.command_data_u.open_form_cmd.formname),last_mod,last_line,"FORM","OPEN");
	  call_cnt +=
	    cache_expression ("", &cmd->cmd_data.command_data_u.open_form_cmd.form_filename, mode);


		sprintf(buff,"<FORM SOURCE=\"%s\" MODULE=\"%s\" LINE=\"%d\">%s</FORM>\n",xml_encode(evaluate_expr(
		cmd->cmd_data.command_data_u.open_form_cmd.form_filename,1
			)) ,
			last_mod,last_line,
			get_form_display( evaluate_expr( cmd->cmd_data.command_data_u.open_form_cmd.form_filename,1))
);


		calltree_addmap(MAPSET_FORM ,buff,last_mod,last_line);
		
		}
	  break;



	case E_CMD_SET_DATABASE_CMD:
	  call_cnt +=
	    cache_expression ("", &cmd->cmd_data.command_data_u.set_database_cmd.set_dbname, mode);
	  break;

	case E_CMD_NEED_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.need_cmd.expr, mode);
	  break;
	case E_CMD_PRINT_CMD:
	  call_cnt +=
	    cache_expression_list ("", cmd->cmd_data.command_data_u.print_cmd.print_expr, mode);
	  break;

	case E_CMD_PRINT_IMG_CMD:
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.print_img_cmd.scale_x, mode);
	  call_cnt += cache_expression ("", &cmd->cmd_data.command_data_u.print_img_cmd.scale_y, mode);
	  break;




	default:
	  break;
	}
    }
return call_cnt;
}


static void dump_variables (variable_list *list, char *type) {
int a;
if (list==0) return;
if (list->variables.variables_len==0) return;

fprintf (output, "<%s_VARIABLES>\n",type);
for (a=0;a<list->variables.variables_len;a++) {
	fprintf(output,"<VARIABLE NAME=\"%s\" LINE=\"%d\"/>\n", list->variables.variables_val[a]->names.names.names_val[0].name, list->variables.variables_val[a]->lineno);
}
fprintf (output, "</%s_VARIABLES>\n",type);
}


int
check_program (module_definition * mods, int nmodules)
{
  //char *dbname = 0;
  int a;
  int b;
  //int *calltree;
  //int bad_load = 0;
  //struct commands *all_cmds = 0;
  char *fname;
  //int mid;

  //load_protos();


  fname = acl_getenv_not_set_as_0 ("CFUNCSFILE");
  if (fname)
    {
      load_boltons (fname);
    }
  else
    {
      printf ("WARNING - not CFUNCSFILE file specified\n");
    }

add_clobberings (mods, nmodules) ;

  for (a = 0; a < nmodules; a++)
    {

      init_whenever ();

      current_mod=&mods[a];

      printf ("Module : %s.4gl\n", mods[a].module_name);
      for (b = 0; b < mods[a].module_entries.module_entries_len; b++)
	{

	  switch (mods[a].module_entries.module_entries_val[b]->met_type)
	    {

	    case E_MET_MAIN_DEFINITION:
	      add_function (a, mods[a].module_name,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.lineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.lastlineno,
			    "MAIN", 'F', &mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition, mods[a].moduleIsInLibrary,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.call_list,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.variables

				);
	      break;

	    case E_MET_FUNCTION_DEFINITION:

	      add_assignments_for_parameters(mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.expanded_parameters, 
				mods[a].module_name,
                            	mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.lineno
			);

	      add_function (a, mods[a].module_name,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.lineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.lastlineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.funcname,
			    'F', &mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition, mods[a].moduleIsInLibrary,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.call_list,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.variables
				);
	      break;

	    case E_MET_REPORT_DEFINITION:
	      add_assignments_for_parameters(mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition.expanded_parameters,
				mods[a].module_name,
                            	mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition.lineno
		);

	      add_function (a, mods[a].module_name,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition.lineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition.lastlineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition.funcname,
			    'R', &mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition, mods[a].moduleIsInLibrary,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.report_definition.call_list,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.variables
		);
	      break;

	    case E_MET_PDF_REPORT_DEFINITION:
	      add_assignments_for_parameters(mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition.expanded_parameters,
				mods[a].module_name,
                            	mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition.lineno
		);
	      add_function (a, mods[a].module_name,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition.lineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition.lastlineno,
			    mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition.funcname,
			    'P', &mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition, mods[a].moduleIsInLibrary,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.pdf_report_definition.call_list,
 				&mods[a].module_entries.module_entries_val[b]->module_entry_u.function_definition.variables
				);
	      break;

	    case E_MET_CMD:
	      {
		struct command *c;
		c = mods[a].module_entries.module_entries_val[b]->module_entry_u.cmd;
		if (c->cmd_data.type == E_CMD_WHENEVER_CMD)
		  {
		    set_whenever (c);
		  }
	      }
	      break;

	    default:
	      // Ignore...
	      break;
	    }
	}

    }


  if (endTarget || 1) {
	printf("Running calltree...");
	run_calltree(top_level_function);
  }

  printf ("Program\n");

  output = fopen ("calltree.xml", "w");

  if (output == 0)
    {
      printf ("Unable to open output file (calltree.xml)\n");
      exit (2);
    }


  if (incProg)
    {
      fprintf (output, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
      fprintf (output, "<PROGRAM>\n");
    }


  if (inc4GL)
    {
      // Dump the sourcecode of the 4gls
      fprintf (output, "<MODULES>\n");
      for (a = 0; a < nmodules; a++)
	{
	  int line;
	  fprintf (output, "<MODULE NAME=\"%s\" MODULENO=\"%d\" FULLNAME=\"%s\" LIBRARY=\"%ld\">\n", mods[a].module_name, a, xml_encode (mods[a].full_path_filename), mods[a].moduleIsInLibrary);

	  dump_variables(&mods[a].exported_global_variables.variables,"GLOBAL");
	  dump_variables(&mods[a].module_variables.variables,"MODULE");

	  // DUMP EXPORTED MODULE VARIABLES
	  // DUMP MODULE VARIABLES

	  for (line = 0; line < mods[a].source_code.lines.lines_len; line++)
	    {
	      fprintf (output, "<LINE>%s</LINE>\n", xml_encode (mods[a].source_code.lines.lines_val[line]));
	    }
	  fprintf (output, "</MODULE>\n");
	}
      fprintf (output, "</MODULES>\n");
    }



  for (a = 0; a < functions_cnt; a++)
    {
	strcpy(currfunc,"_____");
	set_whenever_for_function(&functions[a]);

	if (endTarget) {
		if (functions[a].callsEndTarget==0) { 
			printf("Skipped %s\n",functions[a].function);
			continue; 
		}
	} 

	if (ignore_user_function(functions[a].function)) {	
		continue;
	}

      if (functions[a].f_or_r == 'F')
	{
	  struct s_function_definition *f;
	  strcpy (currfunc, functions[a].function);
	  strcpy (currmod, functions[a].module);

	//printf("Set currfunc to %s\n",currfunc);

	  if (strcmp (functions[a].function, "MAIN") == 0) {
	   	addFunction (currfunc, functions[a].module, functions[a].line,functions[a].lastline, "M");
	  } else {
	        addFunction (currfunc, functions[a].module, functions[a].line,functions[a].lastline, "F");
	  }


	  fprintf (output, "<FUNCTION NAME=\"%s\" TYPE=\"NORMAL\" MODULE=\"%s\"  MODULENO=\"%d\" LINE=\"%d\" LASTLINE=\"%d\" CALLED=\"%d\">\n", 
			functions[a].function, functions[a].module, functions[a].module_no, functions[a].line, functions[a].lastline,functions[a].called);
	dump_variables(functions[a].local_variables,"LOCAL");
	  f = functions[a].ptr;
	  indent++;
	  print_indent ();
	  fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",
					get_first_lineno(f->func_commands ),
					get_last_lineno(f->func_commands )
	  );
	  indent++;
  	  print_whenever (MODE_BUY);
	clr_variable_values();
	printf("Calling add_calltree_calls for %s\n",  functions[a].function);
	  add_calltree_calls ("", f->func_commands, MODE_BUY);
	  indent--;
	  print_indent ();
	  fprintf (output, "</COMMANDS>\n");
	  indent--;
	}
      else
	{
	  int b;
	  struct s_report_definition *r;
	  strcpy (currfunc, functions[a].function);
	  addFunction (currfunc,  functions[a].module, functions[a].line, functions[a].lastline,"R");
	  strcpy (currmod, functions[a].module);



	  if (functions[a].f_or_r == 'P')
	    {
	      fprintf (output, "<FUNCTION NAME=\"%s\" TYPE=\"PDFREPORT\" MODULE=\"%s\" MODULENO=\"%d\" LINE=\"%d\" LASTLINE=\"%d\" CALLED=\"%d\">\n",
		       functions[a].function, functions[a].module, functions[a].module_no, functions[a].line,functions[a].lastline,functions[a].called);
	    }
	  else
	    {
	      fprintf (output, "<FUNCTION NAME=\"%s\" TYPE=\"REPORT\" MODULE=\"%s\" MODULENO=\"%d\" LINE=\"%d\" LASTLINE=\"%d\" CALLED=\"%d\">\n",
		       functions[a].function, functions[a].module, functions[a].module_no, functions[a].line, functions[a].lastline,functions[a].called);
	    }
	  r = functions[a].ptr;
	  indent++;

	  print_indent ();
	  fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n", functions[a].line, functions[a].lastline) ;
  	  print_whenever (MODE_BUY);

	  for (b = 0; b < r->reportFormatSection->entries.entries_len; b++)
	    {
	      if (calls_something (r->reportFormatSection->entries.entries_val[b]->rep_sec_commands))
		{
		  print_indent ();
		  fprintf (output, "<SECTION ID=\"%d\" TYPE=\"%s\" LINE=\"%d\">\n", b,
			   decode_rb (r->reportFormatSection->entries.entries_val[b]->rb_block.rb),
			   r->reportFormatSection->entries.entries_val[b]->lineno);
		  indent++;
		  print_indent ();
		  fprintf (output, "<COMMANDS LINE=\"%d\" LASTLINE=\"%d\">\n",
					get_first_lineno(r->reportFormatSection->entries.entries_val[b]->rep_sec_commands ),
					get_last_lineno(r->reportFormatSection->entries.entries_val[b]->rep_sec_commands )
		  );
		  indent++;
		  add_calltree_calls ("", r->reportFormatSection->entries.entries_val[b]->rep_sec_commands, MODE_BUY);
		  indent--;
		  print_indent ();
		  fprintf (output, "</COMMANDS>\n");
		  indent--;
		  print_indent ();
		  fprintf (output, "</SECTION>\n");
		}
	    }
	  print_indent ();
	  fprintf (output, "</COMMANDS>\n");
	  indent--;
	}


      if (functions[a].f_or_r == 'F')
	{
	  fprintf (output, "</FUNCTION>\n");
	}
      else
	{
	  fprintf (output, "</FUNCTION>\n");
	}

    }

  if (incProg)
    {
	dump_symbols();
  	dump_mapsets();
      fprintf (output, "</PROGRAM>\n");
    }


  printDot ();
  fclose (output);
  return 1;
}



static void
add_bolton (char *fname, char *params, char *rets)
{
  char *p_params[200];
  char *p_rets[200];
  int pcnt = 0;
  int rcnt = 0;
  int l;
  int a;
  nboltons++;

  fprototypes_boltons = realloc (fprototypes_boltons, sizeof (fprototypes_boltons[0]) * nboltons);

  l = strlen (params);

  if (l == 0)
    {
      pcnt = 0;
    }
  else
    {

      pcnt = 0;
      p_params[0] = params;
      for (a = 0; a <= l; a++)
	{
	  if (params[a] == ',' || params[a] == 0)
	    {
	      params[a] = 0;
	      pcnt++;
	      p_params[pcnt] = &params[a + 1];
	      continue;
	    }
	}
    }

  l = strlen (rets);
  if (l == 0)
    {
      rcnt = 0;
    }
  else
    {

      rcnt = 0;
      p_rets[0] = rets;
      for (a = 0; a <= l; a++)
	{
	  if (rets[a] == ',' || rets[a] == 0)
	    {
	      rets[a] = 0;
	      rcnt++;
	      p_rets[rcnt] = &rets[a + 1];
	      continue;
	    }
	}
    }

  fprototypes_boltons[nboltons - 1].pname = strdup (fname);
  fprototypes_boltons[nboltons - 1].proto_type = 0;
  fprototypes_boltons[nboltons - 1].nparams = pcnt;
  if (pcnt)
    {
      fprototypes_boltons[nboltons - 1].param_dtypes = malloc (sizeof (int) * pcnt);
      for (a = 0; a < pcnt; a++)
	{
	  fprototypes_boltons[nboltons - 1].param_dtypes[a] = atoi (p_params[a]);
	}
    }
  else
    {
      fprototypes_boltons[nboltons - 1].param_dtypes = 0;
    }


  fprototypes_boltons[nboltons - 1].nreturns = rcnt;
  if (rcnt)
    {
      fprototypes_boltons[nboltons - 1].return_dtypes = malloc (sizeof (int) * rcnt);
      for (a = 0; a < rcnt; a++)
	{
	  fprototypes_boltons[nboltons - 1].return_dtypes[a] = atoi (p_rets[a]);
	}
    }
  else
    {
      fprototypes_boltons[nboltons - 1].return_dtypes = 0;
    }

}


static void
load_boltons (char *fname)
{
  FILE *f;
  int lineno = 0;
  int printed;
  char buff[2560];
  if (fname == 0)
    return;
  if (strlen (fname) == 0)
    return;
  f = fopen (fname, "r");
  if (f == 0)
    return;

  while (1)
    {
      char *p[4];
      fgets (buff, 2550, f);
      lineno++;
      if (feof (f))
	break;
      A4GL_trim (buff);
      printed = 0;

      p[0] = buff;
      p[1] = strchr (buff, '|');
      if (p[1])
	{
	  *p[1] = 0;
	  p[1]++;
	  p[2] = strchr (p[1], '|');
	  if (p[2])
	    {
	      *p[2] = 0;
	      p[2]++;

		p[3]=strchr(p[2],'|');
		if (p[3]) {
			*p[3]=0;
		}

	      printed++;
	      add_bolton (p[0], p[1], p[2]);
	    }
	}

      if (!printed)
	{
	  printf ("Bad line : %s @ %d\n", buff, lineno);
	  continue;
	}
    }
}


int yylineno;

FILE *lintfile;

module_definition this_module;

int
main (int argc, char *argv[])
{
  module_definition *m;
  int a;
  int b;
init_mapsets();

  if (argc < 2)
    {
      printf ("Usage : %s [-s|-S|-p|-P] infile infile...\n", argv[0]);
      printf (" -s = Fully detailed graph output (includes line numbers)\n");
      printf (" -S = Simple graph output (excludes line numbers) [default]\n");
      printf (" -p = Print only links to internal functions [default]\n");
      printf (" -P = Also print links to external functions\n");
      printf (" -G = Group by module in 'dot' output\n");
      printf (" -Gonly = Group by module in 'dot' output - only include cross-module calls\n");
      printf (" -no4gl = Exclude 4gl sourcecode (required for calltreeviewer)\n");
      printf (" -4gl = Include 4gl sourcecode (required for calltreeviewer) [default]\n");
      printf (" -All = Include all conditionals - whether they call functions or not [default]\n");
      printf (" -NoAll = Exclude conditionals when no calls are made\n");
      printf (" -noProg = Exclude the XML and program tags\n");
      printf (" -Prog = Include the XML and program tags\n");
      printf (" -start funcname = Run calltree from 'funcname'\n");
      printf (" -end funcname = Run calltree for all calls to 'funcname'\n");
      exit (2);
    }

  A4GL_build_user_resources ();
if (A4GL_isyes(acl_getenv("LOGJOINS"))) {
	A4GL_setenv("ALWAYS_CONVERT","Y",1);
	A4GL_setenv("SETTABLEFORCOL","Y",1);
}


  ignore_user_function(NULL);
  m = malloc (sizeof (struct module_definition) * (argc - 1));


  a = 0;
  for (b = 1; b < argc; b++)
    {
      char buff[256];
      if (strcmp (argv[b], "-S") == 0)
	{
	  simpleGraph = 1;
	  continue;
	}
      if (strcmp (argv[b], "-All") == 0)
	{
	  AlwaysIncludeAllConditionals = 1;
	  continue;
	}
      if (strcmp (argv[b], "-NoAll") == 0)
	{
	  AlwaysIncludeAllConditionals = 0;
	  continue;
	}

      if (strcmp (argv[b], "-s") == 0)
	{
	  simpleGraph = 0;
	  continue;
	}
      if (strcmp (argv[b], "-no4gl") == 0)
	{
	  inc4GL = 0;
	  continue;
	}
      if (strcmp (argv[b], "-G") == 0) 
	{
		groupByModule=1;
		continue;
	}

      if (strcmp (argv[b], "-Gonly") == 0) 
	{
		groupByModule=2;
		continue;
	}

      if (strcmp (argv[b], "-4gl") == 0)
	{
	  inc4GL = 1;
	  continue;
	}
      if (strcmp (argv[b], "-lib") == 0)
	{
	   ignLibFunc = 0;
	  continue;
	}

      if (strcmp (argv[b], "-nolib") == 0)
	{
	   ignLibFunc = 1;
	  continue;
	}

      if (strcmp (argv[b], "-P") == 0)
	{
	  printAllFuncs = 1;
	  continue;
	}

      if (strcmp (argv[b], "-p") == 0)
	{
	  printAllFuncs = 0;
	  continue;
	}

      if (strcmp (argv[b], "-Prog") == 0)
	{
	  incProg = 1;
	  continue;
	}

      if (strcmp (argv[b], "-start") == 0) {
		b++;
		strcpy(top_level_function,argv[b]);
		continue;
	}

      if (strcmp (argv[b], "-end") == 0) {
		b++;
		endTarget=strdup(argv[b]);
		continue;
	}

      if (strcmp (argv[b], "-NoProg") == 0)
	{
	  incProg = 0;
	  continue;
	}

      a++;
      sprintf (buff, "%s", argv[b]);
      if (strstr (buff, ".dat") != 0)
	{
	  char *p;
	  p = strstr (buff, ".dat");
	  *p = 0;
	}
      printf ("Loading %s : ", buff);
      fflush (stdout);
      if (read_module_definition (&m[a - 1], buff))
	{
	  printf ("OK...%s\n", argv[b]);
	}
      else
	{
	  printf ("- Failed to load %s\n", argv[b]);
	  exit (1);
	}
    }

  check_program (m, a);
  dump_mapsets();

  return 0;
}

char *
decode_rb (enum report_blocks a)
{
  switch (a)
    {
    case RB_FIRST_PAGE_HEADER:
      return "FIRST_PAGE_HEADER";
    case RB_PAGE_HEADER:
      return "PAGE_HEADER";
    case RB_PAGE_TRAILER:
      return "PAGE_TRAILER";
    case RB_ON_EVERY_ROW:
      return "ON_EVERY_ROW";
    case RB_ON_LAST_ROW:
      return "ON_LAST_ROW";

    case RB_FORMAT_EVERY_ROW:
      return "FORMAT_EVERY_ROW";

    case RB_BEFORE_GROUP_OF:
      return "BEFORE_GROUP_OF";
    case RB_AFTER_GROUP_OF:
      return "AFTER_GROUP_OF";
    }

  return "Blah3";
}





static int
run_calltree_for (s_call_list * call_list)
{
  int a;
  int cnt = 0;
  struct s_expr_function_call *fcall;

  for (a = 0; a < call_list->calls_by_expr.calls_by_expr_len; a++)
    {
      switch (call_list->calls_by_expr.calls_by_expr_val[a]->expr_type)
	{
	case ET_EXPR_FCALL:
	  fcall = call_list->calls_by_expr.calls_by_expr_val[a]->expr_str_u.  expr_function_call;
	  cnt += run_calltree (fcall->functionname);
	  break;

	default:
	  break;
	}
    }

  for (a = 0; a < call_list->calls_by_call.calls_by_call_len; a++)
    {
					command_ptr cmd=call_list->calls_by_call.calls_by_call_val[a];
      switch ( cmd->cmd_data.type) {
			case E_CMD_CALL_CMD: // Should be covered by the expression in the call above..
					break;

			case E_CMD_START_CMD: 
	  			cnt += run_calltree (cmd->cmd_data.command_data_u.start_cmd.repname);
				break;

			case E_CMD_OUTPUT_CMD: 
	  			cnt += run_calltree (cmd->cmd_data.command_data_u.output_cmd.repname);
				break;

			case E_CMD_FINISH_CMD: 
	  			cnt += run_calltree (cmd->cmd_data.command_data_u.finish_cmd.repname);
				break;
						
			default :
					printf("Unexpected command in calls_by_call list");
					break;

	  break;
	}
	}


  return cnt;
}


// Go through and generate the whitelist of all functions
// which are actually called...
static int run_calltree (char *s)
{
  int a;
  int cnt=0;


printf("Run calltree : %s\n",s);

  if (endTarget) {
	if (strcmp(s,endTarget)==0) {
		//printf("Run_calltree for %s is endTarget\n",s);
		cnt++;
	}
  }

  if (system_function (s))
    {
      return 0;
    }

  for (a = 0; a < functions_cnt; a++)
    {
      if (strcmp (functions[a].function, s) == 0)
	{
	  if (functions[a].called == 0)
	    {
	      functions[a].called = 1;
			printf("Running calltree for ...\n");
	      cnt+=run_calltree_for (functions[a].call_list);
		if (cnt) {
			functions[a].callsEndTarget=1;
		}
	    }
	  //printf("Run_calltree for %s returns %d\n", s, cnt);
	  return cnt;
	}
    }
  return 0;
}



static char * evaluate_variable_usage_as_string(variable_usage *var_usage) {
        char buff[2000];


        sprintf(buff, "%s",var_usage->variable_name);
        if (var_usage->subscripts.subscripts_len) {
                int a;
                strcat(buff, "[");
                for (a=0;a<var_usage->subscripts.subscripts_len;a++) {
                        if(a) strcat(buff, ",");
                        strcat(buff, evaluate_expr(var_usage->subscripts.subscripts_val[a],1));
                }
                strcat(buff, "]");
        }
        if (var_usage->substrings_start.substrings_start) {
                strcat(buff, "[");
                strcat(buff, evaluate_expr(var_usage->substrings_start.substrings_start,1));
                if (var_usage->substrings_end.substrings_end) {
                        strcat(buff, ",");
                        strcat(buff, evaluate_expr(var_usage->substrings_end.substrings_end,1));
                }
                strcat(buff, "]");
        }
        if (var_usage->next) {
                char *ptr;
                strcat(buff,".");
                ptr=evaluate_variable_usage_as_string(var_usage->next);
                strcat(buff, ptr);
                free(ptr);
        }

        return strdup(buff);
}


struct variable_values {
	char *name;
	char *value;
	int nassignments;
} *variable_values=0;
int nvariable_values=0;


static void clr_variable_values(void) {
	int a;
	for (a=0;a<nvariable_values;a++) {
		if (variable_values[a].name) free(variable_values[a].name);
		if (variable_values[a].value) free(variable_values[a].value);
	}
	free(variable_values);
	variable_values=0;
	nvariable_values=0;
}



static void inc_inIf(void) {
	inIf++;
	ifBlocks[inIf]=0;
}

static void dec_inIf(void) {
/*
	char mybuff[200000]="";
	int a;
	char *ptr;

	for (a=0;a<nvariable_values;a++) {
		char varbuff[2000];
		sprintf(varbuff,"IF_%d_",inIf);
		if (A4GL_strstartswith(variable_values[a].name,varbuff)) {
			variable_values[a].name=0;
			if (strlen(mybuff)==0) {
				sprintf(mybuff,"[%s", variable_values[a].value);
			} else {
				strcat(mybuff,"|");
				strcat(mybuff, variable_values[a].value);
			}
			free(variable_values[a].value);
			variable_values[a].value=0;
		}
	}
*/

	inIf--;
}


static void add_variable_value(variable_usage *u, expr_str *val) {
	char *value;
	char *var;
	//int a;
	var=evaluate_variable_usage_as_string(u);

	if (u->escope!=E_SCOPE_LOCAL && u->escope!=E_SCOPE_MODULE) {
		//fprintf(stderr, "Ignoring nonlocal variable .. %s\n",var);
		return;
	}

	value=evaluate_expr(val,0);

  	if (inIf && A4GL_isyes(acl_getenv("TRYTOMARKIFVAREXPRS")) ) {
		static char buff[10000];
		//static char varbuff[10000];
		sprintf(buff,"(%s)",value);
		value=buff;
		//sprintf(varbuff,"IF_%d_%s",inIf,var);
		//var=varbuff;
	}

	add_variable_value_strings(var,value);
}



static void add_variable_value_strings (char *var, char *value) {
int a;
	int gap=-1;


//printf("%s=%s Ifcnt=%d\n",var,value,inIf);
	for (a=0;a<nvariable_values;a++) {

		if (variable_values[a].name==0) {
			if (gap==-1) gap=a;
			continue;
		}
		if (strcmp(variable_values[a].name,var)==0) {
			free(variable_values[a].value);
			variable_values[a].value=strdup(value);
			variable_values[a].nassignments++;
			return ;
		}
	}


	if (gap==-1) {
	 	nvariable_values++;

		variable_values=realloc(variable_values, sizeof(variable_values[0])*nvariable_values);
	 	variable_values[nvariable_values-1].name=strdup(var);
	 	variable_values[nvariable_values-1].value=strdup(value);
		variable_values[nvariable_values-1].nassignments=1;
	} else {
 		variable_values[gap].name=strdup(var);
 		variable_values[gap].value=strdup(value);
		variable_values[gap].nassignments=1;
	}
	
	//printf("var=%s ",var);
	//printf("value=%s\n",value);
}


static char *variable_value_string (char *var,int useName) {
	int a;
	for (a=0;a<nvariable_values;a++) {
		if (variable_values[a].name==0) continue;
		if (strcmp(variable_values[a].name, var)==0) {
			return variable_values[a].value;
		}
	}

	if (strcmp(var,"gfrmdir")==0) { return "$FGLFORM"; }

        if (strcmp(var,"glogname")==0) {return "$LOGNAME";}
        if (strcmp(var,"gytty")==0) { return "$YTTY";}
        if (strcmp(var,"g4gldir")==0) { return "$FGLDIR";}
        if (strcmp(var,"gfrmdir")==0) { return "$FGLFORM";}

        if (strcmp(var,"gpid")==0) { return "$PID";}

        if (strcmp(var,"gtempfile")==0) { return "$FGLTMP/temp$YTTY";}
        if (strcmp(var,"greportfile")==0) { return "$FGLTMP/report$YTTY";}
        if (strcmp(var,"gtemptable")==0) { return "temp$YTTY";}

	if (useName) {
		static char buff[200];
		sprintf(buff,"{%s}",var);
		return buff;
	} else {
		return "?";
	}
}



static char *variable_value(variable_usage *u,int useName) {
	char *var;
	var=evaluate_variable_usage_as_string(u);
	
	return variable_value_string(var,useName);
}



/*
static int has_variable_value(variable_usage *u) {
	if (variable_value(u,0)) return TRUE;
	return FALSE;
}*/





/* UseName == Use variable name if no value is known - otherwise - use ? */
char *
evaluate_expr (expr_str * e,int useName)
{
if (e==0) return "";

//printf("e->expr_type=%d\n",e->expr_type);
//fflush(stdout);
  switch (e->expr_type)
    {
    case ET_EXPR_LITERAL_EMPTY_STRING:
      return "";
      //case ET_EXPR_SUBSTRING: return "?1?";
    case ET_EXPR_LITERAL_STRING:
      {
	if ( e->expr_str_u.expr_string) {
		return strdup(e->expr_str_u.expr_string);
	} else {
		return "";
		}
      }
      break;


    case ET_EXPR_VARIABLE_USAGE:
      return strdup (variable_value (e->expr_str_u.expr_variable_usage,useName));
      break;

    case ET_EXPR_BRACKET:
      {
	char *ptr;
	ptr = strdup (evaluate_expr (e->expr_str_u.expr_expr,useName));
	return strdup (ptr);
      }
      break;


    case ET_EXPR_SQLERRMESSAGE:
      return strdup ("SQLERRMESSAGE");
      break;

    case ET_EXPR_LITERAL_LONG:

      {
	static char smbuff[200];
	sprintf (smbuff, "%ld", e->expr_str_u.expr_long);
	return strdup (smbuff);
      }
      break;

    case ET_EXPR_OP_CLIP:

      {
	char *p;
	p = strdup (evaluate_expr (e->expr_str_u.expr_expr,useName));
	A4GL_trim (p);
	return p;
      }
      break;


    case ET_EXPR_FALSE:
      return "0";

    case ET_EXPR_TRUE:
      return "1";

    case ET_EXPR_GET_FLDBUF:
      return "<GET_FLDBUF>";

    case ET_EXPR_ASCII:
      {
	char buff[20000];
	char *ptr;
	ptr = strdup (evaluate_expr (e->expr_str_u.expr_expr,useName));
	if (strcmp (ptr, "?"))
	  {
	    return "?";
	  }

	sprintf (buff, "%d", ptr[0]);
	free (ptr);
	return strdup (buff);
      }

    case ET_EXPR_OP_CONCAT:
      {
	char buff[20000];
	//printf("CONCAT %s\n",evaluate_expr (e->expr_str_u.expr_op->left));
	//printf(" +  %s\n",evaluate_expr (e->expr_str_u.expr_op->right));
	sprintf (buff, "%s%s", evaluate_expr (e->expr_str_u.expr_op->left,useName), evaluate_expr (e->expr_str_u.expr_op->right,useName));
	return strdup (buff);
      }

    case ET_EXPR_OP_USING:
      {
	//char buff[20000];
	//char fmt[2000];
	//char left[2000];
		
		return evaluate_expr (e->expr_str_u.expr_op->left,1);

	/*
	if (strcmp(evaluate_expr (e->expr_str_u.expr_op->left,0),"?")==0) {
		char buff[2000];
		sprintf(buff,"%s",expr_as_string_when_possible(e->expr_str_u.expr_op->left),  evaluate_expr (e->expr_str_u.expr_op->right,0));
		return strdup(buff);
	} else {
	}
		*/
	//return strdup ("USING");
      }
      break;

    case ET_EXPR_NOT:
      {
	if (atol (evaluate_expr (e->expr_str_u.expr_expr,useName)))
	  {
	    return strdup ("0");
	  }
	else
	  {
	    return strdup ("1");
	  }
      }
      break;

    case ET_EXPR_NULL:
      return strdup ("");

    case ET_EXPR_LITERAL_DOUBLE_STR:
      {
	char buff[256];
	sprintf (buff, "%s", e->expr_str_u.expr_string);
	return strdup (buff);
      }

    case ET_EXPR_TIME_EXPR:
      return "TIME";

    case ET_EXPR_TODAY:
      return "TODAY";


    case ET_EXPR_DAY_FUNC:
      {
	char buff[256];
	sprintf (buff, "DAY(%s)", evaluate_expr (e->expr_str_u.expr_expr,useName));
	return strdup (buff);
      }
      break;

    case ET_EXPR_YEAR_FUNC:
      {
	char buff[256];
	sprintf (buff, "YEAR(%s)", evaluate_expr (e->expr_str_u.expr_expr,useName));
	return strdup (buff);
      }
      break;

    case ET_EXPR_UPSHIFT:
      {
	char *p;
	p = strdup (evaluate_expr (e->expr_str_u.expr_expr,useName));
	upshift (p);
	return p;
      }
      break;

    case ET_EXPR_DOWNSHIFT:
      {
	char *p;
	p = strdup (evaluate_expr (e->expr_str_u.expr_expr,useName));
	downshift (p);
	return p;
      }
      break;

    case ET_EXPR_CAST:
      {
	char buff[20000];
	sprintf (buff, "%s", evaluate_expr (e->expr_str_u.expr_cast->expr,useName));
	return strdup (buff);
      }
      break;

    case ET_E_V_OR_LIT_STRING:
      return strdup (e->expr_str_u.expr_string);

  case ET_EXPR_FCALL:
    
	  if (strcmp(e->expr_str_u.expr_function_call->functionname,"fgl_getenv")==0) {
		char buff[2000];
		char *ptr;
		ptr=evaluate_expr(e->expr_str_u.expr_function_call->parameters->list.list_val[0],useName);
		//printf("ptr=%s\n",ptr);
		if (acl_getenv_not_set_as_0(ptr)) {
			return strdup(acl_getenv_not_set_as_0(ptr));
		}
		sprintf(buff,"$%s",ptr);
		return strdup(buff);
	  } else {
	  	return "?";
	  }
	  break;
	

    default:
      //A4GL_pause_execution ();
      return "?";

    }

}


/*
FILE *crudfile=0;


static void ensure_crud(void) {
if (crudfile==0) {
	crudfile=fopen("crud.out","w");
}
}
*/


static void add_table_action(char *mainaction, char *tabname, char *action,char *module,int line) {
char buff[200000];

if (strstr(action,"SELECT")) {
sprintf(buff,"<SELECT_STMT TABLE=\"%s\" MODULE=\"%s\" LINE=\"%d\" MAINACTION=\"%s\"/>\n",tabname,module,line, mainaction);
	calltree_addmap(MAPSET_CRUD_SELECT, buff,module,line);
	return;
}
if (strstr(action,"UPDATE")) {
if (strcmp(tabname,"tdetails")==0) {
	printf("-----------------------------\n");
}
sprintf(buff,"<UPDATE_STMT TABLE=\"%s\" MODULE=\"%s\" LINE=\"%d\" MAINACTION=\"%s\"/>\n",tabname,module,line, mainaction);
	calltree_addmap(MAPSET_CRUD_UPDATE, buff,module,line);
	return;
}
if (strstr(action,"INSERT")) {
sprintf(buff,"<INSERT_STMT TABLE=\"%s\" MODULE=\"%s\" LINE=\"%d\" MAINACTION=\"%s\"/>\n",tabname,module,line, mainaction);
	calltree_addmap(MAPSET_CRUD_INSERT, buff,module,line);
	return;
}
if (strstr(action,"DELETE")) {
sprintf(buff,"<DELETE_STMT TABLE=\"%s\" MODULE=\"%s\" LINE=\"%d\" MAINACTION=\"%s\"/>\n",tabname,module,line, mainaction);
	calltree_addmap(MAPSET_CRUD_DELETE, buff,module,line);
	return;
}
sprintf(buff,"<STATEMENT TABLE=\"%s\" ACTION=\"%s\" MODULE=\"%s\" LINE=\"%d\" MAINACTION=\"%s\"/>\n",tabname,action,module,line, mainaction);

calltree_addmap(MAPSET_CRUD_OTHER, buff,module,line);
}



void
calltree_map_select_stmt (char *main_statement_type, struct s_select *select, char *module, int line)
{
  
    int a;
    //char *into_temp = 0;

    if (select->table_elements.tables.tables_len > 10000)
      {
	A4GL_assertion (1, "Dubious number of tables!");
      }


    A4GLSQLPARSE_from_clause_collect_tables (select, select->first, &select->table_elements);

    for (a = 0; a < select->table_elements.tables.tables_len; a++)
      {
	char *tabname;
	char *alias;
	tabname = strdup (select->table_elements.tables.tables_val[a].tabname);
	alias = select->table_elements.tables.tables_val[a].alias;
	
	if (alias == 0 )
	  {
	    alias = tabname;
	  }

	if (strlen(alias)==0) {
	    alias = tabname;
	}
        alias = strdup (alias);
        A4GL_trim (alias);
	
	A4GL_trim (tabname);
	//add_table_action(main_statement_type, alias, "SELECT",module,line);
	add_table_action(main_statement_type, tabname, "SELECT",module,line);
      }


    calltree_map_value_list (main_statement_type, select->select_list, module, line);
	if (select->where_clause) {
    		calltree_map_value (main_statement_type, select->where_clause, module, line);
	}
    if (select->next)
      {
	calltree_map_select_stmt (main_statement_type, select->next,module,line);
      }

}

static void calltree_map_value(char *main_statement,struct s_select_list_item *item, char*module, int line) {
	if (item==0) return;
//printf("In calltree_map_value : %s %d\n",main_statement, item->data.type);

	switch (item->data.type) {
        case E_SLI_IBIND:
        case E_SLI_VARIABLE:
        case E_SLI_DATETIME:
        case E_SLI_INTERVAL:
        case E_SLI_COLUMN_NOT_TRANSFORMED:
        case E_SLI_LITERAL:
        case E_SLI_CHAR: 
        case E_SLI_VAR_REPLACE:
        case E_SLI_COLUMN:
        case E_SLI_COLUMN_ORDERBY:
        case E_SLI_VARIABLE_USAGE_IN_SELECT_LIST :
		break;


        case E_SLI_VARIABLE_USAGE_LIST:
		{
		int a;
		struct s_select_list_item_list *vl;
		vl=item->data.s_select_list_item_data_u.var_usage_list;
		for (a=0;a<vl->list.list_len;a++)  {
			calltree_map_value(main_statement,vl->list.list_val[a], module, line);
		}
		}
		break;

        case E_SLI_VARIABLE_USAGE:
		{
		struct expr_str e;
		e.expr_type=ET_EXPR_VARIABLE_USAGE;	
		e.expr_str_u.expr_variable_usage=item->data.s_select_list_item_data_u.var_usage;
		add_variable(&e,module,line,"USE",NULL);
		}
		break;

        case E_SLI_BUILTIN_CONST_COUNT_STAR:
        case E_SLI_BUILTIN_CONST_CURRENT:
        case E_SLI_BUILTIN_CONST_FALSE:
        case E_SLI_BUILTIN_CONST_ROWID:
        case E_SLI_BUILTIN_CONST_STAR:
        case E_SLI_BUILTIN_CONST_TIME:
        case E_SLI_BUILTIN_CONST_TODAY:
        case E_SLI_BUILTIN_CONST_TRUE:
        case E_SLI_BUILTIN_CONST_USER:
        case E_SLI_BUILTIN_CONST_NULL:
        case E_SLI_QUERY_PLACEHOLDER:
		break;


        case E_SLI_CAST_EXPR:
		calltree_map_value(main_statement, item->data.s_select_list_item_data_u.casting.expr,module,line);
		break;

        case E_SLI_ASC:
        case E_SLI_DESC:
        case E_SLI_NOT:
        case E_SLI_BRACKET_EXPR:
        case E_SLI_UNITS_YEAR:
        case E_SLI_UNITS_DAY:
        case E_SLI_UNITS_HOUR:
        case E_SLI_UNITS_MINUTE:
        case E_SLI_UNITS_MONTH:
        case E_SLI_UNITS_SECOND:
        case E_SLI_ISNOTNULL:
        case E_SLI_ISNULL:
		calltree_map_value(main_statement, item->data.s_select_list_item_data_u.expr,module,line);
		break;


      	case E_SLI_BUILTIN_FUNC_DATE:
        case E_SLI_BUILTIN_FUNC_DAY:
        case E_SLI_BUILTIN_FUNC_DOW:
        case E_SLI_BUILTIN_FUNC_MDY:
        case E_SLI_BUILTIN_FUNC_MONTH:
        case E_SLI_BUILTIN_FUNC_WEEKDAY:
        case E_SLI_BUILTIN_FUNC_YEAR:
		calltree_map_value_list(main_statement,item->data.s_select_list_item_data_u.builtin_fcall.params,module,line);
		break;

        case E_SLI_BUILTIN_AGG_AVG:
        case E_SLI_BUILTIN_AGG_COUNT:
        case E_SLI_BUILTIN_AGG_MAX:
        case E_SLI_BUILTIN_AGG_MIN:
        case E_SLI_BUILTIN_AGG_SUM:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.agg_expr.expr,module,line);
		break;

        case E_SLI_EXTEND:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.extend.expr,module,line);
		break;

       case E_SLI_FCALL:
		calltree_map_value_list(main_statement,item->data.s_select_list_item_data_u.fcall.params,module,line);
		break;

        case E_SLI_CASE:
		{
			int a;
			for (a=0;a<item->data.s_select_list_item_data_u.sqlcase.elements.elements_len;a++) {
				calltree_map_value(main_statement,item->data.s_select_list_item_data_u.sqlcase.elements.elements_val[a],module,line);
			}
		}
		break;



        case E_SLI_REGEX_ILIKE:
        case E_SLI_REGEX_LIKE:
        case E_SLI_REGEX_MATCHES:
        case E_SLI_REGEX_NOT_ILIKE:
        case E_SLI_REGEX_NOT_LIKE:
        case E_SLI_REGEX_NOT_MATCHES:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.regex.val,module,line);
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.regex.regex,module,line);
	break;

        case E_SLI_BETWEEN:
        case E_SLI_NOT_BETWEEN:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.between_expr.val,module,line);
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.between_expr.from,module,line);
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.between_expr.to,module,line);
	break;

        case E_SLI_CASE_ELEMENT:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.sqlcaseelement.condition,module,line);
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.sqlcaseelement.response,module,line);
		break;

        case E_SLI_JOIN:
        case E_SLI_OP:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.complex_expr.left,module,line);
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.complex_expr.right,module,line);
		
		break;

        case E_SLI_IN_SELECT:
        case E_SLI_NOT_IN_SELECT:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.sli_expr.left,module,line);
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.sli_expr.right,module,line);
		break;


        case E_SLI_IN_VALUES:
        case E_SLI_NOT_IN_VALUES:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.slil_expr.left,module,line);
		calltree_map_value_list(main_statement,item->data.s_select_list_item_data_u.slil_expr.right_list,module,line);
		break;

        case E_SLI_QUERY:
        case E_SLI_SUBQUERY:
		calltree_map_select_stmt(main_statement, item->data.s_select_list_item_data_u.subquery, module,line);
		break;

        case E_SLI_SUBQUERY_EXPRESSION:
		calltree_map_value(main_statement,item->data.s_select_list_item_data_u.sq_expression.sq,module,line);
		break;


	}

}

void calltree_map_value_list(char *main_statement,struct s_select_list_item_list *list, char*module, int line) {
int a;
if (list==0) return;
for (a=0;a<list->list.list_len;a++) {
	struct s_select_list_item *si;
	si=list->list.list_val[a];
	calltree_map_value(main_statement,si,module,line);
}

}


void calltree_map_insert_delete_update(char *main_statement, char *table, struct expr_str *where_clause, char *module, int line) {
	add_table_action(main_statement, table,main_statement,module,line);
	if (where_clause) {
		switch (where_clause->expr_type) {
			case ET_EXPR_SELECT_LIST_ITEM:
				{
				struct s_select_list_item *sl;
				sl=where_clause->expr_str_u.sl_item;
				calltree_map_value(main_statement,sl,module,line);
				} 
			break;

			case ET_EXPR_WHERE_CURRENT_OF:
				return;

			default:
				A4GL_assertion(1,"Unexpected WHERE CLAUSE expression type");
			}
	}
}




static void calltree_map_sql(char *buff,char *module, int line) {
	char buffx[200000];
	sprintf(buffx, "<SQL MODULE=\"%s\" LINE=\"%d\" STMT=\"%s\" />\n",module,line, xml_encode(buff));
	calltree_addmap(MAPSET_CRUD_OTHER, buffx,module,line);
}

static char * guess_sql_stmt(struct struct_prepare_cmd *p,char *module, int line,int processed) {
static char buff[200000];
static char buff2[200000];
char *ptr;
	strcpy(buff,evaluate_expr(p->sql,0));

	ptr=calltree_get_ident(p->stmtid); 

	sprintf(buff2, "<PREPARE STMTID=\"%s\" MODULE=\"%s\" LINE=\"%d\" STMT=\"%s\" PROCESSED=\"%d\"/>\n",ptr, module,line, xml_encode(buff), processed);
	calltree_addmap(MAPSET_CRUD_DYNAMIC,buff2,module,line);
	return buff;
}




static void calltree_addmap (enum e_mapset mapset, char *fmt, char *module, int line) {
int a;
//va_list args;
//char buff[200025];
//va_start (args, fmt);

//vsprintf (buff, fmt, args);
//A4GL_assertion(strlen(buff)>=sizeof(buff),"Buffer too small");

for (a=0;a<map_crud[mapset].nlines;a++) {
	if (strcmp(map_crud[mapset].lines[a],fmt)==0) {
		return ;
	}
}

if (isLineCalled(module,line)) {
	map_crud[mapset].nlines++;
	map_crud[mapset].lines=realloc(map_crud[mapset].lines,map_crud[mapset].nlines*sizeof(char *));
	map_crud[mapset].lines[map_crud[mapset].nlines-1]=strdup(fmt);
}
}


static char *basename(char *s) {
char *p;
p=strrchr(s,'/');
if (p) {
	p++; 
	return p;
}
return s;
}


static int isLineCalled(char *module, int line) {
	int a;
	for (a=0;a<functions_cnt;a++) {
		if (strcmp(basename(functions[a].module), basename(module))==0) {
			if (functions[a].line<=line && functions[a].lastline>=line) {
				if (functions[a].called) {
					return 1;
				} else {
					return 0;
				}
			}
		}
	}
	//printf("Line %s - %d is not called\n",module,line);
return 0;
}

char *expand_env_vars_in_cmdline(char *str,int show_errors);

static void add_assignments_for_parameters(expr_str_list *params,char *module,int lineno) {
int a;
if (params==0) return;
for (a=0;a<params->list.list_len;a++) {
	add_variable(params->list.list_val[a],module,lineno,"PARAM",NULL);
}
}


static char *get_form_display(char *s) {
	// This function returns what a form looks like
	// this should be in a file with the parameter ".out" 
	// eg. if a form is called 'bibble' - we try to load 'bibble.out'
	// If the file is not found - we just return ""
	//
	// This file should normally be just the 'screen' section.
	// You can generate one using the "fdecompile_scr" command.
	// This is not compiled by default - you'll need to compile it
	// in compilers/fcompile
	//
//char *ptr;
static char buff[20000];
char strbuff[2000];
//char bp[2000];
char *b;
FILE *f;


b=buff;
strcpy(b,expand_env_vars_in_cmdline(s,0));
if (b[0]=='(') {
	b++;
	b[strlen(b)-1]=0;
}
sprintf(buff,"%s.out",b);
printf("Looking for form : %s\n",buff);
f=(FILE *) A4GL_open_file_dbpath (buff);
if (f==0) {
	//printf("Not found\n");
	return "";
}
//printf("FOund!!\n");
strcpy(buff,"<![CDATA[\n");
while (1) {
	fgets(strbuff,255,f);
		if (feof(f)) break;
	strcat(buff,strbuff);
}
strcat(buff,"\n]]>");
//printf("buff=%s\n",buff);
return buff;
}



static void  dump_parameters(struct s_expr_function_call *fcall,int filenameParamNo) {
//char buff[3000];
/*
	printf("%s(",fcall->functionname);
	if (fcall->parameters) {
		int a;
		for (a=0;a<fcall->parameters->list.list_len;a++) {
			if (a) printf(",");
			printf("%s",evaluate_expr(fcall->parameters->list.list_val[a]));
		}
	}
	printf(") @%s.%d\n",fcall->module,fcall->line);
*/

	add_file_operation( fcall->functionname, fcall->parameters->list.list_val[filenameParamNo], fcall->module,fcall->line);

}


static void add_file_operation(char *operation, expr_str *file, char *module,int line) {
		char buff[2000];

		sprintf(buff,"<FILE OPERATION=\"%s\" FILENAME=\"%s\" MODULE=\"%s\" LINE=\"%d\"/>\n",
			operation,
			xml_encode(evaluate_expr(file,1)), module,line);

		calltree_addmap(MAPSET_FILES ,buff,module,line);
}

//
// A place to log calls to specific application specific functions - 
// this will depend on the site...
//
static void log_special_calls(struct s_expr_function_call *fcall) {
char *Fname;
Fname=fcall->functionname;

if (strcmp(Fname,"fgl_file_exist")==0) {  dump_parameters(fcall,0); }
if (strcmp(Fname,"fgl_write_str")==0) {  dump_parameters(fcall,1);}
if (strcmp(Fname,"fgl_directory_exist")==0) {  dump_parameters(fcall,0); }
if (strcmp(Fname,"fgl_open_write")==0) {  dump_parameters(fcall,0); }
if (strcmp(Fname,"fgl_get_file_size")==0) {  dump_parameters(fcall,0); }
//if (strcmp(Fname,"fgl_write_line")==0) {  dump_parameters(fcall); }
//if (strcmp(Fname,"write_cdf_file")==0) {  dump_parameters(fcall); }

if (strcmp(Fname,"fgl_open_read")==0) {  dump_parameters(fcall,0); }
if (strcmp(Fname,"fgl_write_fixed_str")==0) {  dump_parameters(fcall,1); }
if (strcmp(Fname,"fgl_delete_file")==0) {  dump_parameters(fcall,0); }
if (strcmp(Fname,"fgl_write_list_str")==0) {  dump_parameters(fcall,1); }


}




static int get_first_lineno(struct s_commands *c) {
	if (c) {
		return c->first_lineno;
	} 
	return -1;
}

static int get_last_lineno(struct s_commands *c) {
	if (c) {
		return c->last_lineno;
	} 
	return -1;
}
